import dataclasses
import json
import logging
import re
from datetime import datetime
from typing import Any, Dict, List, Optional, Tuple
from urllib.parse import urlparse

import dateutil.parser
import requests
from packaging import version
from pydantic import BaseModel, Field, validator

from datahub.configuration.git import GitReference
from datahub.configuration.validate_field_rename import pydantic_renamed_field
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.source.aws.aws_common import AwsConnectionConfig
from datahub.ingestion.source.aws.s3_util import is_s3_uri
from datahub.ingestion.source.dbt.dbt_common import (
    DBTColumn,
    DBTCommonConfig,
    DBTModelPerformance,
    DBTNode,
    DBTSourceBase,
    DBTSourceReport,
)
from datahub.ingestion.source.dbt.dbt_tests import DBTTest, DBTTestResult

logger = logging.getLogger(__name__)


@dataclasses.dataclass
class DBTCoreReport(DBTSourceReport):
    catalog_info: Optional[dict] = None
    manifest_info: Optional[dict] = None


class DBTCoreConfig(DBTCommonConfig):
    manifest_path: str = Field(
        description="Path to dbt manifest JSON. See https://docs.getdbt.com/reference/artifacts/manifest-json. "
        "This can be a local file or a URI."
    )
    catalog_path: Optional[str] = Field(
        None,
        description="Path to dbt catalog JSON. See https://docs.getdbt.com/reference/artifacts/catalog-json. "
        "This file is optional, but highly recommended. Without it, some metadata like column info will be incomplete or missing. "
        "This can be a local file or a URI.",
    )
    sources_path: Optional[str] = Field(
        default=None,
        description="Path to dbt sources JSON. See https://docs.getdbt.com/reference/artifacts/sources-json. "
        "If not specified, last-modified fields will not be populated. "
        "This can be a local file or a URI.",
    )
    run_results_paths: List[str] = Field(
        default=[],
        description="Path to output of dbt test run as run_results files in JSON format. "
        "If not specified, test execution results and model performance metadata will not be populated in DataHub."
        "If invoking dbt multiple times, you can provide paths to multiple run result files. "
        "See https://docs.getdbt.com/reference/artifacts/run-results-json.",
    )

    only_include_if_in_catalog: bool = Field(
        default=False,
        description="[experimental] If true, only include nodes that are also present in the catalog file. "
        "This is useful if you only want to include models that have been built by the associated run.",
    )

    # Because we now also collect model performance metadata, the "test_results" field was renamed to "run_results".
    _convert_test_results_path = pydantic_renamed_field(
        "test_results_path", "run_results_paths", transform=lambda x: [x] if x else []
    )
    _convert_run_result_path_singular = pydantic_renamed_field(
        "run_results_path", "run_results_paths", transform=lambda x: [x] if x else []
    )

    aws_connection: Optional[AwsConnectionConfig] = Field(
        default=None,
        description="When fetching manifest files from s3, configuration for aws connection details",
    )

    git_info: Optional[GitReference] = Field(
        None,
        description="Reference to your git location to enable easy navigation from DataHub to your dbt files.",
    )

    _github_info_deprecated = pydantic_renamed_field("github_info", "git_info")

    @validator("aws_connection", always=True)
    def aws_connection_needed_if_s3_uris_present(
        cls, aws_connection: Optional[AwsConnectionConfig], values: Dict, **kwargs: Any
    ) -> Optional[AwsConnectionConfig]:
        # first check if there are fields that contain s3 uris
        uris = [
            values.get(f)
            for f in [
                "manifest_path",
                "catalog_path",
                "sources_path",
            ]
        ] + values.get("run_results_paths", [])
        s3_uris = [uri for uri in uris if is_s3_uri(uri or "")]

        if s3_uris and aws_connection is None:
            raise ValueError(
                f"Please provide aws_connection configuration, since s3 uris have been provided {s3_uris}"
            )
        return aws_connection


def get_columns(
    dbt_name: str,
    catalog_node: Optional[dict],
    manifest_node: dict,
    tag_prefix: str,
) -> List[DBTColumn]:
    manifest_columns = manifest_node.get("columns", {})
    manifest_columns_lower = {k.lower(): v for k, v in manifest_columns.items()}

    if catalog_node is not None:
        logger.debug(f"Loading schema info for {dbt_name}")
        catalog_columns = catalog_node["columns"]
    elif manifest_columns:
        # If the end user ran `dbt compile` instead of `dbt docs generate`, then the catalog
        # file will not have any column information. In this case, we will fall back to using
        # information from the manifest file.
        logger.debug(f"Inferring schema info for {dbt_name} from manifest")
        catalog_columns = {
            k: {"name": col["name"], "type": col["data_type"] or "", "index": i}
            for i, (k, col) in enumerate(manifest_columns.items())
        }
    else:
        logger.debug(f"Missing schema info for {dbt_name}")
        return []

    columns = []
    for key, catalog_column in catalog_columns.items():
        manifest_column = manifest_columns.get(
            key, manifest_columns_lower.get(key.lower(), {})
        )

        meta = manifest_column.get("meta", {})

        tags = manifest_column.get("tags", [])
        tags = [tag_prefix + tag for tag in tags]

        dbtCol = DBTColumn(
            name=catalog_column["name"],
            comment=catalog_column.get("comment", ""),
            description=manifest_column.get("description", ""),
            data_type=catalog_column["type"],
            index=catalog_column["index"],
            meta=meta,
            tags=tags,
        )
        columns.append(dbtCol)
    return columns


def extract_dbt_entities(
    all_manifest_entities: Dict[str, Dict[str, Any]],
    all_catalog_entities: Optional[Dict[str, Dict[str, Any]]],
    sources_results: List[Dict[str, Any]],
    manifest_adapter: str,
    use_identifiers: bool,
    tag_prefix: str,
    only_include_if_in_catalog: bool,
    include_database_name: bool,
    report: DBTSourceReport,
) -> List[DBTNode]:
    sources_by_id = {x["unique_id"]: x for x in sources_results}

    dbt_entities = []
    for key, manifest_node in all_manifest_entities.items():
        name = manifest_node["name"]

        if use_identifiers and manifest_node.get("identifier"):
            name = manifest_node["identifier"]

        if (
            manifest_node.get("alias") is not None
            and manifest_node.get("resource_type")
            != "test"  # tests have non-human-friendly aliases, so we don't want to use it for tests
        ):
            name = manifest_node["alias"]

        materialization = None
        if "materialized" in manifest_node.get("config", {}):
            # It's a model
            materialization = manifest_node["config"]["materialized"]

        upstream_nodes = []
        if "depends_on" in manifest_node and "nodes" in manifest_node["depends_on"]:
            upstream_nodes = manifest_node["depends_on"]["nodes"]

        catalog_node = (
            all_catalog_entities.get(key) if all_catalog_entities is not None else None
        )
        missing_from_catalog = catalog_node is None
        catalog_type = None

        if catalog_node is None:
            if materialization in {"test", "ephemeral"}:
                # Test and ephemeral nodes will never show up in the catalog.
                missing_from_catalog = False
            else:
                if all_catalog_entities is not None and not only_include_if_in_catalog:
                    # If the catalog file is missing, we have already generated a general message.
                    report.warning(
                        title="Node missing from catalog",
                        message="Found a node in the manifest file but not in the catalog. "
                        "This usually means the catalog file was not generated by `dbt docs generate` and so is incomplete. "
                        "Some metadata, particularly schema information, will be impacted.",
                        context=key,
                    )
        else:
            catalog_type = catalog_node["metadata"]["type"]

        # initialize comment to "" for consistency with descriptions
        # (since dbt null/undefined descriptions as "")
        comment = ""
        if catalog_node is not None and catalog_node.get("metadata", {}).get("comment"):
            comment = catalog_node["metadata"]["comment"]

        query_tag_props = manifest_node.get("query_tag", {})

        meta = manifest_node.get("meta", {})

        owner = meta.get("owner")
        if owner is None:
            owner = (manifest_node.get("config", {}).get("meta") or {}).get("owner")

        if not meta:
            # On older versions of dbt, the meta field was nested under config
            # for some node types.
            meta = manifest_node.get("config", {}).get("meta") or {}

        tags = manifest_node.get("tags", [])
        tags = [tag_prefix + tag for tag in tags]

        max_loaded_at_str = sources_by_id.get(key, {}).get("max_loaded_at")
        max_loaded_at = None
        if max_loaded_at_str:
            max_loaded_at = parse_dbt_timestamp(max_loaded_at_str)

        test_info = None
        if manifest_node.get("resource_type") == "test":
            test_metadata = manifest_node.get("test_metadata", {})
            kw_args = test_metadata.get("kwargs", {})

            qualified_test_name = (
                (test_metadata.get("namespace") or "")
                + "."
                + (test_metadata.get("name") or "")
            )
            qualified_test_name = (
                qualified_test_name[1:]
                if qualified_test_name.startswith(".")
                else qualified_test_name
            )
            test_info = DBTTest(
                qualified_test_name=qualified_test_name,
                column_name=kw_args.get("column_name"),
                kw_args=kw_args,
            )

        dbtNode = DBTNode(
            dbt_name=key,
            dbt_adapter=manifest_adapter,
            dbt_package_name=manifest_node.get("package_name"),
            database=manifest_node["database"] if include_database_name else None,
            schema=manifest_node["schema"],
            name=name,
            alias=manifest_node.get("alias"),
            dbt_file_path=manifest_node["original_file_path"],
            node_type=manifest_node["resource_type"],
            max_loaded_at=max_loaded_at,
            comment=comment,
            description=manifest_node.get("description", ""),
            raw_code=manifest_node.get(
                "raw_code", manifest_node.get("raw_sql")
            ),  # Backward compatibility dbt <=v1.2
            language=manifest_node.get(
                "language", "sql"
            ),  # Backward compatibility dbt <=v1.2
            upstream_nodes=upstream_nodes,
            materialization=materialization,
            catalog_type=catalog_type,
            missing_from_catalog=missing_from_catalog,
            meta=meta,
            query_tag=query_tag_props,
            tags=tags,
            owner=owner,
            compiled_code=manifest_node.get(
                "compiled_code", manifest_node.get("compiled_sql")
            ),  # Backward compatibility dbt <=v1.2
            test_info=test_info,
        )

        # Load columns from catalog, and override some properties from manifest.
        if dbtNode.materialization not in [
            "ephemeral",
            "test",
        ]:
            dbtNode.columns = get_columns(
                dbtNode.dbt_name,
                catalog_node,
                manifest_node,
                tag_prefix,
            )

        else:
            dbtNode.columns = []

        dbt_entities.append(dbtNode)

    return dbt_entities


def parse_dbt_timestamp(timestamp: str) -> datetime:
    return dateutil.parser.parse(timestamp)


class DBTRunTiming(BaseModel):
    name: Optional[str] = None
    started_at: Optional[str] = None
    completed_at: Optional[str] = None

    # TODO parse these into datetime objects


class DBTRunResult(BaseModel):
    class Config:
        extra = "allow"

    status: str
    timing: List[DBTRunTiming] = []
    unique_id: str
    failures: Optional[int] = None
    message: Optional[str] = None

    @property
    def timing_map(self) -> Dict[str, DBTRunTiming]:
        return {x.name: x for x in self.timing if x.name}

    def has_success_status(self) -> bool:
        return self.status in ("pass", "success")


class DBTRunMetadata(BaseModel):
    dbt_schema_version: str
    dbt_version: str
    generated_at: str
    invocation_id: str


def _parse_test_result(
    dbt_metadata: DBTRunMetadata,
    run_result: DBTRunResult,
) -> Optional[DBTTestResult]:
    if not run_result.has_success_status():
        native_results = {"message": run_result.message or ""}
        if run_result.failures:
            native_results.update({"failures": str(run_result.failures)})
    else:
        native_results = {}

    execution_timestamp = run_result.timing_map.get("execute")
    if execution_timestamp and execution_timestamp.started_at:
        execution_timestamp_parsed = parse_dbt_timestamp(execution_timestamp.started_at)
    else:
        execution_timestamp_parsed = parse_dbt_timestamp(dbt_metadata.generated_at)

    return DBTTestResult(
        invocation_id=dbt_metadata.invocation_id,
        status=run_result.status,
        native_results=native_results,
        execution_time=execution_timestamp_parsed,
    )


def _parse_model_run(
    dbt_metadata: DBTRunMetadata,
    run_result: DBTRunResult,
) -> Optional[DBTModelPerformance]:
    status = run_result.status
    if status not in {"success", "error"}:
        return None

    execution_timestamp = run_result.timing_map.get("execute")
    if not execution_timestamp:
        return None
    if not execution_timestamp.started_at or not execution_timestamp.completed_at:
        return None

    return DBTModelPerformance(
        run_id=dbt_metadata.invocation_id,
        status=status,
        start_time=parse_dbt_timestamp(execution_timestamp.started_at),
        end_time=parse_dbt_timestamp(execution_timestamp.completed_at),
    )


def load_run_results(
    config: DBTCommonConfig,
    test_results_json: Dict[str, Any],
    all_nodes: List[DBTNode],
) -> List[DBTNode]:
    if test_results_json.get("args", {}).get("which") == "generate":
        logger.warning(
            "The run results file is from a `dbt docs generate` command, "
            "instead of a build/run/test command. Skipping this file."
        )
        return all_nodes

    dbt_metadata = DBTRunMetadata.parse_obj(test_results_json.get("metadata", {}))

    all_nodes_map: Dict[str, DBTNode] = {x.dbt_name: x for x in all_nodes}

    results = test_results_json.get("results", [])
    for result in results:
        run_result = DBTRunResult.parse_obj(result)
        id = run_result.unique_id

        if id.startswith("test."):
            test_result = _parse_test_result(dbt_metadata, run_result)
            if not test_result:
                continue

            test_node = all_nodes_map.get(id)
            if not test_node:
                logger.debug(f"Failed to find test node {id} in the catalog")
                continue

            assert test_node.test_info is not None
            test_node.test_results.append(test_result)

        else:
            model_performance = _parse_model_run(dbt_metadata, run_result)
            if not model_performance:
                continue

            model_node = all_nodes_map.get(id)
            if not model_node:
                logger.debug(f"Failed to find model node {id} in the catalog")
                continue

            model_node.model_performances.append(model_performance)

    return all_nodes


@platform_name("dbt")
@config_class(DBTCoreConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.TEST_CONNECTION, "Enabled by default")
class DBTCoreSource(DBTSourceBase, TestableSource):
    config: DBTCoreConfig
    report: DBTCoreReport

    def __init__(self, config: DBTCommonConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.report = DBTCoreReport()

    @classmethod
    def create(cls, config_dict, ctx):
        config = DBTCoreConfig.parse_obj(config_dict)
        return cls(config, ctx)

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        test_report = TestConnectionReport()
        try:
            source_config = DBTCoreConfig.parse_obj_allow_extras(config_dict)
            DBTCoreSource.load_file_as_json(
                source_config.manifest_path, source_config.aws_connection
            )
            if source_config.catalog_path is not None:
                DBTCoreSource.load_file_as_json(
                    source_config.catalog_path, source_config.aws_connection
                )
            test_report.basic_connectivity = CapabilityReport(capable=True)
        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=str(e)
            )
        return test_report

    @staticmethod
    def load_file_as_json(
        uri: str, aws_connection: Optional[AwsConnectionConfig]
    ) -> Dict:
        if re.match("^https?://", uri):
            return json.loads(requests.get(uri).text)
        elif is_s3_uri(uri):
            u = urlparse(uri)
            assert aws_connection
            response = aws_connection.get_s3_client().get_object(
                Bucket=u.netloc, Key=u.path.lstrip("/")
            )
            return json.loads(response["Body"].read().decode("utf-8"))
        else:
            with open(uri) as f:
                return json.load(f)

    def loadManifestAndCatalog(
        self,
    ) -> Tuple[
        List[DBTNode],
        Optional[str],
        Optional[str],
        Optional[str],
        Optional[str],
        Optional[str],
    ]:
        dbt_manifest_json = self.load_file_as_json(
            self.config.manifest_path, self.config.aws_connection
        )
        dbt_manifest_metadata = dbt_manifest_json["metadata"]
        self.report.manifest_info = dict(
            generated_at=dbt_manifest_metadata.get("generated_at", "unknown"),
            dbt_version=dbt_manifest_metadata.get("dbt_version", "unknown"),
            project_name=dbt_manifest_metadata.get("project_name", "unknown"),
        )

        dbt_catalog_json = None
        dbt_catalog_metadata = None
        if self.config.catalog_path is not None:
            dbt_catalog_json = self.load_file_as_json(
                self.config.catalog_path, self.config.aws_connection
            )
            dbt_catalog_metadata = dbt_catalog_json.get("metadata", {})
            self.report.catalog_info = dict(
                generated_at=dbt_catalog_metadata.get("generated_at", "unknown"),
                dbt_version=dbt_catalog_metadata.get("dbt_version", "unknown"),
                project_name=dbt_catalog_metadata.get("project_name", "unknown"),
            )
        else:
            self.report.warning(
                title="No catalog file configured",
                message="Some metadata, particularly schema information, will be missing.",
            )

        if self.config.sources_path is not None:
            dbt_sources_json = self.load_file_as_json(
                self.config.sources_path, self.config.aws_connection
            )
            sources_results = dbt_sources_json["results"]
        else:
            sources_results = {}

        manifest_schema = dbt_manifest_json["metadata"].get("dbt_schema_version")
        manifest_version = dbt_manifest_json["metadata"].get("dbt_version")
        manifest_adapter = dbt_manifest_json["metadata"].get("adapter_type")

        catalog_schema = None
        catalog_version = None
        if dbt_catalog_metadata is not None:
            catalog_schema = dbt_catalog_metadata.get("dbt_schema_version")
            catalog_version = dbt_catalog_metadata.get("dbt_version")

        manifest_nodes = dbt_manifest_json["nodes"]
        manifest_sources = dbt_manifest_json["sources"]

        all_manifest_entities = {**manifest_nodes, **manifest_sources}

        all_catalog_entities = None
        if dbt_catalog_json is not None:
            catalog_nodes = dbt_catalog_json["nodes"]
            catalog_sources = dbt_catalog_json["sources"]

            all_catalog_entities = {**catalog_nodes, **catalog_sources}

        nodes = extract_dbt_entities(
            all_manifest_entities=all_manifest_entities,
            all_catalog_entities=all_catalog_entities,
            sources_results=sources_results,
            manifest_adapter=manifest_adapter,
            use_identifiers=self.config.use_identifiers,
            tag_prefix=self.config.tag_prefix,
            only_include_if_in_catalog=self.config.only_include_if_in_catalog,
            include_database_name=self.config.include_database_name,
            report=self.report,
        )

        return (
            nodes,
            manifest_schema,
            manifest_version,
            manifest_adapter,
            catalog_schema,
            catalog_version,
        )

    def load_nodes(self) -> Tuple[List[DBTNode], Dict[str, Optional[str]]]:
        (
            all_nodes,
            manifest_schema,
            manifest_version,
            manifest_adapter,
            catalog_schema,
            catalog_version,
        ) = self.loadManifestAndCatalog()

        # If catalog_version is between 1.7.0 and 1.7.2, report a warning.
        try:
            if (
                catalog_version
                and catalog_version.startswith("1.7.")
                and version.parse(catalog_version) < version.parse("1.7.3")
            ):
                self.report.report_warning(
                    title="Dbt Catalog Version",
                    message="Due to a bug in dbt version between 1.7.0 and 1.7.2, you will have incomplete metadata "
                    "source",
                    context=f"Due to a bug in dbt, dbt version {catalog_version} will have incomplete metadata on "
                    f"sources."
                    "Please upgrade to dbt version 1.7.3 or later. "
                    "See https://github.com/dbt-labs/dbt-core/issues/9119 for details on the bug.",
                )
        except Exception as e:
            self.report.info(
                title="dbt Catalog Version",
                message="Failed to determine the catalog version",
                exc=e,
            )

        additional_custom_props = {
            "manifest_schema": manifest_schema,
            "manifest_version": manifest_version,
            "manifest_adapter": manifest_adapter,
            "catalog_schema": catalog_schema,
            "catalog_version": catalog_version,
        }

        for run_results_path in self.config.run_results_paths:
            # This will populate the test_results and model_performance fields on each node.
            all_nodes = load_run_results(
                self.config,
                self.load_file_as_json(run_results_path, self.config.aws_connection),
                all_nodes,
            )

        return all_nodes, additional_custom_props

    def _filter_nodes(self, all_nodes: List[DBTNode]) -> List[DBTNode]:
        nodes = super()._filter_nodes(all_nodes)

        if not self.config.only_include_if_in_catalog:
            return nodes

        filtered_nodes = []
        for node in nodes:
            if node.missing_from_catalog:
                # TODO: We need to do some additional testing of this flag to validate that it doesn't
                # drop important things entirely (e.g. sources).
                self.report.nodes_filtered.append(node.dbt_name)
            else:
                filtered_nodes.append(node)

        return filtered_nodes

    def get_external_url(self, node: DBTNode) -> Optional[str]:
        if self.config.git_info and node.dbt_file_path:
            return self.config.git_info.get_url_for_file_path(node.dbt_file_path)
        return None



import logging
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass, field
from datetime import datetime, timezone
from functools import partial
from typing import Dict, Iterable, List, Optional

from pydantic import Field

from datahub.configuration import ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceReport
from datahub.ingestion.api.source_helpers import auto_workunit_reporter
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.graph.client import DataHubGraph
from datahub.utilities.lossy_collections import LossyList
from datahub.utilities.stats_collections import TopKDict

logger = logging.getLogger(__name__)

DATAJOB_QUERY = """
query listDataJobs($query:String!, $scrollId: String, $batchSize: Int) {
  scrollAcrossEntities(input: { types: [DATA_JOB
    ], query: $query, count: $batchSize, scrollId: $scrollId
  }) {
    nextScrollId
    count
    searchResults {
      entity {
        type
        ... on DataJob {
          urn
          dataFlow {
            urn
          }
          type
          lastIngested
          subTypes {
            typeNames
          }
          jobId
          dataPlatformInstance {
            urn
          }
         runs {
            total
          }
        }
      }
    }
  }
}
"""

DATAFLOW_QUERY = """
query listDataFlows($query:String!, $scrollId: String, $batchSize: Int) {
  scrollAcrossEntities(input: { types: [DATA_FLOW], query: $query, count: $batchSize, scrollId: $scrollId}) {
    nextScrollId
    count
    searchResults {
      entity {
        type
        ... on DataFlow {
            urn
            type
            orchestrator
            cluster
            lastIngested
        }
      }
    }
  }
}
"""


DATA_PROCESS_INSTANCES_QUERY = """
query getDataJobRuns($dataJobUrn: String!, $start: Int!, $count: Int!) {
  dataJob(urn: $dataJobUrn) {
    runs(start: $start, count: $count) {
      runs {
        created {
          time
          actor
        }
        urn
      }
      #...runResult

      __typename
    }
    __typename
  }
}
"""


class DataProcessCleanupConfig(ConfigModel):
    enabled: bool = Field(
        default=True, description="Whether to do data process cleanup."
    )
    retention_days: Optional[int] = Field(
        10,
        description="Number of days to retain metadata in DataHub",
    )

    aspects_to_clean: List[str] = Field(
        ["DataprocessInstance"],
        description="List of aspect names to clean up",
    )

    keep_last_n: Optional[int] = Field(
        5,
        description="Number of latest aspects to keep",
    )

    delete_empty_data_jobs: bool = Field(
        False, description="Whether to delete Data Jobs without runs"
    )

    delete_empty_data_flows: bool = Field(
        False, description="Whether to delete Data Flows without runs"
    )

    hard_delete_entities: bool = Field(
        False,
        description="Whether to hard delete entities",
    )

    batch_size: int = Field(
        500,
        description="The number of entities to get in a batch from API",
    )

    max_workers: int = Field(
        10,
        description="The number of workers to use for deletion",
    )

    delay: Optional[float] = Field(
        0.25,
        description="Delay between each batch",
    )


@dataclass
class DataFlowEntity:
    urn: str
    orchestrator: Optional[str]
    cluster: Optional[str]
    last_ingested: Optional[int]


@dataclass
class DataJobEntity:
    urn: str
    flow_urn: str
    lastIngested: Optional[int]
    jobId: Optional[str]
    dataPlatformInstance: Optional[str]
    total_runs: int = 0


@dataclass
class DataProcessCleanupReport(SourceReport):
    num_aspects_removed: int = 0
    num_aspect_removed_by_type: TopKDict[str, int] = field(default_factory=TopKDict)
    sample_soft_deleted_aspects_by_type: TopKDict[str, LossyList[str]] = field(
        default_factory=TopKDict
    )
    num_data_flows_found: int = 0
    num_data_jobs_found: int = 0


class DataProcessCleanup:
    """
    This source is a maintenance source which cleans up old/unused aspects.

    Currently it only supports:
        - DataFlow
        - DataJob
        - DataProcessInstance

    """

    def __init__(
        self,
        ctx: PipelineContext,
        config: DataProcessCleanupConfig,
        report: DataProcessCleanupReport,
        dry_run: bool = False,
    ):
        if not ctx.graph:
            raise ValueError("MetadataCleanupSource needs a datahub_api")

        self.graph: DataHubGraph = ctx.graph
        self.ctx = ctx
        self.config = config
        self.report = report
        self.dry_run = dry_run

    def get_report(self) -> DataProcessCleanupReport:
        return self.report

    # auto_work_unit_report is overriden to disable a couple of automation like auto status aspect, etc.. which is not needed her.
    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [partial(auto_workunit_reporter, self.get_report())]

    def fetch_dpis(self, job_urn: str, batch_size: int) -> List[dict]:
        assert self.ctx.graph
        dpis = []
        start = 0
        # This graphql endpoint doesn't support scrolling and therefore after 10k DPIs it causes performance issues on ES
        # Therefore, we are limiting the max DPIs to 9000
        max_item = 9000
        while True:
            try:
                job_query_result = self.ctx.graph.execute_graphql(
                    DATA_PROCESS_INSTANCES_QUERY,
                    {"dataJobUrn": job_urn, "start": start, "count": batch_size},
                )
                job_data = job_query_result.get("dataJob")
                if not job_data:
                    logger.error(f"Error getting job {job_urn}")
                    break

                runs_data = job_data.get("runs")
                if not runs_data:
                    logger.error(f"Error getting runs for {job_urn}")
                    break

                runs = runs_data.get("runs")
                dpis.extend(runs)
                start += batch_size
                if len(runs) < batch_size or start >= max_item:
                    break
            except Exception as e:
                self.report.failure(
                    f"Exception while fetching DPIs for job {job_urn}:", exc=e
                )
                break
        return dpis

    def keep_last_n_dpi(
        self, dpis: List[Dict], job: DataJobEntity, executor: ThreadPoolExecutor
    ) -> None:
        if not self.config.keep_last_n:
            return

        deleted_count_last_n = 0
        if len(dpis) >= self.config.keep_last_n:
            futures = {}
            for dpi in dpis[self.config.keep_last_n :]:
                future = executor.submit(
                    self.delete_entity, dpi["urn"], "dataprocessInstance"
                )
                futures[future] = dpi

            for future in as_completed(futures):
                try:
                    future.result()
                    deleted_count_last_n += 1
                    futures[future]["deleted"] = True
                except Exception as e:
                    self.report.report_failure(
                        f"Exception while deleting DPI: {e}", exc=e
                    )
            if (
                deleted_count_last_n % self.config.batch_size == 0
                and deleted_count_last_n > 0
            ):
                logger.info(f"Deleted {deleted_count_last_n} DPIs from {job.urn}")
                if self.config.delay:
                    logger.info(f"Sleeping for {self.config.delay} seconds")
                    time.sleep(self.config.delay)

        if deleted_count_last_n > 0:
            logger.info(f"Deleted {deleted_count_last_n} DPIs from {job.urn}")

    def delete_entity(self, urn: str, type: str) -> None:
        assert self.ctx.graph

        self.report.num_aspects_removed += 1
        self.report.num_aspect_removed_by_type[type] = (
            self.report.num_aspect_removed_by_type.get(type, 0) + 1
        )
        if type not in self.report.sample_soft_deleted_aspects_by_type:
            self.report.sample_soft_deleted_aspects_by_type[type] = LossyList()
        self.report.sample_soft_deleted_aspects_by_type[type].append(urn)

        if self.dry_run:
            logger.info(
                f"Dry run is on otherwise it would have deleted {urn} with hard deletion is {self.config.hard_delete_entities}"
            )
            return

        self.ctx.graph.delete_entity(urn, self.config.hard_delete_entities)

    def delete_dpi_from_datajobs(self, job: DataJobEntity) -> None:
        assert self.ctx.graph

        dpis = self.fetch_dpis(job.urn, self.config.batch_size)
        dpis.sort(
            key=lambda x: x["created"]["time"]
            if x.get("created") and x["created"].get("time")
            else 0,
            reverse=True,
        )

        with ThreadPoolExecutor(max_workers=self.config.max_workers) as executor:
            if self.config.keep_last_n:
                self.keep_last_n_dpi(dpis, job, executor)

            if self.config.retention_days is not None:
                self.remove_old_dpis(dpis, job, executor)

        job.total_runs = len(
            list(
                filter(lambda dpi: "deleted" not in dpi or not dpi.get("deleted"), dpis)
            )
        )

    def remove_old_dpis(
        self, dpis: List[Dict], job: DataJobEntity, executor: ThreadPoolExecutor
    ) -> None:
        if self.config.retention_days is None:
            return

        deleted_count_retention = 0
        retention_time = (
            int(datetime.now(timezone.utc).timestamp())
            - self.config.retention_days * 24 * 60 * 60
        )

        futures = {}
        for dpi in dpis:
            if dpi.get("deleted"):
                continue

            if (
                not dpi.get("created")
                or not dpi["created"].get("time")
                or dpi["created"]["time"] < retention_time * 1000
            ):
                future = executor.submit(
                    self.delete_entity, dpi["urn"], "dataprocessInstance"
                )
                futures[future] = dpi

        for future in as_completed(futures):
            try:
                future.result()
                deleted_count_retention += 1
                futures[future]["deleted"] = True
            except Exception as e:
                self.report.report_failure(f"Exception while deleting DPI: {e}", exc=e)

            if (
                deleted_count_retention % self.config.batch_size == 0
                and deleted_count_retention > 0
            ):
                logger.info(
                    f"Deleted {deleted_count_retention} DPIs from {job.urn} due to retention"
                )

                if self.config.delay:
                    logger.info(f"Sleeping for {self.config.delay} seconds")
                    time.sleep(self.config.delay)

        if deleted_count_retention > 0:
            logger.info(
                f"Deleted {deleted_count_retention} DPIs from {job.urn} due to retention"
            )
        else:
            logger.debug(f"No DPIs to delete from {job.urn} due to retention")

    def get_data_flows(self) -> Iterable[DataFlowEntity]:
        assert self.ctx.graph

        scroll_id: Optional[str] = None
        previous_scroll_id: Optional[str] = None

        while True:
            result = None
            try:
                result = self.ctx.graph.execute_graphql(
                    DATAFLOW_QUERY,
                    {
                        "query": "*",
                        "scrollId": scroll_id if scroll_id else None,
                        "batchSize": self.config.batch_size,
                    },
                )
            except Exception as e:
                self.report.failure(
                    f"While trying to get dataflows with {scroll_id}", exc=e
                )
                break

            scrollAcrossEntities = result.get("scrollAcrossEntities")
            if not scrollAcrossEntities:
                raise ValueError("Missing scrollAcrossEntities in response")
            self.report.num_data_flows_found += scrollAcrossEntities.get("count")
            logger.info(f"Got {scrollAcrossEntities.get('count')} DataFlow entities")

            scroll_id = scrollAcrossEntities.get("nextScrollId")
            for flow in scrollAcrossEntities.get("searchResults"):
                yield DataFlowEntity(
                    urn=flow.get("entity").get("urn"),
                    orchestrator=flow.get("entity").get("orchestrator"),
                    cluster=flow.get("entity").get("cluster"),
                    last_ingested=flow.get("entity").get("lastIngested"),
                )

            if not scroll_id or previous_scroll_id == scroll_id:
                break

            previous_scroll_id = scroll_id

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        if not self.config.enabled:
            return []
        assert self.ctx.graph

        dataFlows: Dict[str, DataFlowEntity] = {}
        if self.config.delete_empty_data_flows:
            for flow in self.get_data_flows():
                dataFlows[flow.urn] = flow

        scroll_id: Optional[str] = None
        previous_scroll_id: Optional[str] = None

        dataJobs: Dict[str, List[DataJobEntity]] = defaultdict(list)
        deleted_jobs: int = 0

        while True:
            try:
                result = self.ctx.graph.execute_graphql(
                    DATAJOB_QUERY,
                    {
                        "query": "*",
                        "scrollId": scroll_id if scroll_id else None,
                        "batchSize": self.config.batch_size,
                    },
                )
            except Exception as e:
                self.report.failure(
                    f"While trying to get data jobs with {scroll_id}", exc=e
                )
                break
            scrollAcrossEntities = result.get("scrollAcrossEntities")
            if not scrollAcrossEntities:
                raise ValueError("Missing scrollAcrossEntities in response")

            self.report.num_data_jobs_found += scrollAcrossEntities.get("count")
            logger.info(f"Got {scrollAcrossEntities.get('count')} DataJob entities")

            scroll_id = scrollAcrossEntities.get("nextScrollId")
            for job in scrollAcrossEntities.get("searchResults"):
                datajob_entity = DataJobEntity(
                    urn=job.get("entity").get("urn"),
                    flow_urn=job.get("entity").get("dataFlow").get("urn"),
                    lastIngested=job.get("entity").get("lastIngested"),
                    jobId=job.get("entity").get("jobId"),
                    dataPlatformInstance=job.get("entity").get("dataPlatformInstance"),
                    total_runs=job.get("entity").get("runs").get("total"),
                )
                if datajob_entity.total_runs > 0:
                    try:
                        self.delete_dpi_from_datajobs(datajob_entity)
                    except Exception as e:
                        self.report.failure(
                            f"While trying to delete {datajob_entity} ", exc=e
                        )
                if (
                    datajob_entity.total_runs == 0
                    and self.config.delete_empty_data_jobs
                ):
                    logger.info(
                        f"Deleting datajob {datajob_entity.urn} because there are no runs"
                    )
                    self.delete_entity(datajob_entity.urn, "dataJob")
                    deleted_jobs += 1
                    if deleted_jobs % self.config.batch_size == 0:
                        logger.info(f"Deleted {deleted_jobs} DataJobs")
                else:
                    dataJobs[datajob_entity.flow_urn].append(datajob_entity)

            if not scroll_id or previous_scroll_id == scroll_id:
                break

            previous_scroll_id = scroll_id

        if deleted_jobs > 0:
            logger.info(f"Deleted {deleted_jobs} DataJobs")
        # Delete empty dataflows if needed
        if self.config.delete_empty_data_flows:
            deleted_data_flows: int = 0
            for key in dataFlows:
                if not dataJobs.get(key) or len(dataJobs[key]) == 0:
                    logger.info(
                        f"Deleting dataflow {key} because there are not datajobs"
                    )
                    self.delete_entity(key, "dataFlow")
                    deleted_data_flows += 1
                    if deleted_jobs % self.config.batch_size == 0:
                        logger.info(f"Deleted {deleted_data_flows} DataFlows")
            logger.info(f"Deleted {deleted_data_flows} DataFlows")

        return []



import logging
import os
from typing import Any, Dict, Optional

from google.api_core.client_info import ClientInfo
from google.cloud import bigquery, datacatalog_v1, resourcemanager_v3
from google.cloud.logging_v2.client import Client as GCPLoggingClient
from pydantic import Field, PrivateAttr

from datahub._version import __version__
from datahub.configuration.common import ConfigModel
from datahub.ingestion.source.common.gcp_credentials_config import GCPCredential

logger = logging.getLogger(__name__)


def _get_bigquery_client_info() -> ClientInfo:
    """Get ClientInfo with DataHub user-agent for BigQuery client identification"""
    return ClientInfo(user_agent=f"datahub/{__version__}")


class BigQueryConnectionConfig(ConfigModel):
    credential: Optional[GCPCredential] = Field(
        default=None, description="BigQuery credential informations"
    )

    _credentials_path: Optional[str] = PrivateAttr(None)

    extra_client_options: Dict[str, Any] = Field(
        default={},
        description="Additional options to pass to google.cloud.logging_v2.client.Client.",
    )

    project_on_behalf: Optional[str] = Field(
        default=None,
        description="[Advanced] The BigQuery project in which queries are executed. Will be passed when creating a job. If not passed, falls back to the project associated with the service account.",
    )

    def __init__(self, **data: Any):
        super().__init__(**data)

        if self.credential:
            self._credentials_path = self.credential.create_credential_temp_file()
            logger.debug(
                f"Creating temporary credential file at {self._credentials_path}"
            )
            os.environ["GOOGLE_APPLICATION_CREDENTIALS"] = self._credentials_path

    def get_bigquery_client(self) -> bigquery.Client:
        client_options = self.extra_client_options
        return bigquery.Client(
            self.project_on_behalf,
            client_info=_get_bigquery_client_info(),
            **client_options,
        )

    def get_projects_client(self) -> resourcemanager_v3.ProjectsClient:
        return resourcemanager_v3.ProjectsClient()

    def get_policy_tag_manager_client(self) -> datacatalog_v1.PolicyTagManagerClient:
        return datacatalog_v1.PolicyTagManagerClient()

    def make_gcp_logging_client(
        self, project_id: Optional[str] = None
    ) -> GCPLoggingClient:
        # See https://github.com/googleapis/google-cloud-python/issues/2674 for
        # why we disable gRPC here.
        client_options = self.extra_client_options.copy()
        client_options["_use_grpc"] = False
        if project_id is not None:
            return GCPLoggingClient(**client_options, project=project_id)
        else:
            return GCPLoggingClient(**client_options)

    def get_sql_alchemy_url(self) -> str:
        if self.project_on_behalf:
            return f"bigquery://{self.project_on_behalf}"
        # When project_id is not set, we will attempt to detect the project ID
        # based on the credentials or environment variables.
        # See https://github.com/mxmzdlv/pybigquery#authentication.
        return "bigquery://"



import logging
import sys
from typing import Any, Dict, List, Optional
from urllib.parse import quote

import requests

from datahub.ingestion.source.qlik_sense.config import Constant, QlikSourceConfig
from datahub.ingestion.source.qlik_sense.data_classes import (
    PERSONAL_SPACE_DICT,
    App,
    Chart,
    Item,
    QlikDataset,
    QlikTable,
    Sheet,
    Space,
)
from datahub.ingestion.source.qlik_sense.websocket_connection import WebsocketConnection

# Logger instance
logger = logging.getLogger(__name__)


class QlikAPI:
    def __init__(self, config: QlikSourceConfig) -> None:
        self.spaces: Dict = {}
        self.users: Dict = {}
        self.config = config
        self.session = requests.Session()
        self.session.headers.update(
            {
                "Authorization": f"Bearer {self.config.api_key}",
                "Content-Type": "application/json",
            }
        )
        self.rest_api_url = f"https://{self.config.tenant_hostname}/api/v1"
        # Test connection by fetching list of api keys
        logger.info(f"Trying to connect to {self.rest_api_url}")
        self.session.get(f"{self.rest_api_url}/api-keys").raise_for_status()

    def _log_http_error(self, message: str) -> Any:
        logger.warning(message)
        _, e, _ = sys.exc_info()
        if isinstance(e, requests.exceptions.HTTPError):
            logger.warning(f"HTTP status-code = {e.response.status_code}")
        logger.debug(msg=message, exc_info=e)
        return e

    def get_spaces(self) -> List[Space]:
        spaces: List[Space] = []
        try:
            url = f"{self.rest_api_url}/spaces"
            while True:
                response = self.session.get(url)
                response.raise_for_status()
                response_dict = response.json()
                for space_dict in response_dict[Constant.DATA]:
                    space = Space.parse_obj(space_dict)
                    spaces.append(space)
                    self.spaces[space.id] = space.name
                if Constant.NEXT in response_dict[Constant.LINKS]:
                    url = response_dict[Constant.LINKS][Constant.NEXT][Constant.HREF]
                else:
                    break
            # Add personal space entity
            spaces.append(Space.parse_obj(PERSONAL_SPACE_DICT))
            self.spaces[PERSONAL_SPACE_DICT[Constant.ID]] = PERSONAL_SPACE_DICT[
                Constant.NAME
            ]
        except Exception as e:
            self._log_http_error(message=f"Unable to fetch spaces. Exception: {e}")
        return spaces

    def _get_dataset(self, dataset_id: str, item_id: str) -> Optional[QlikDataset]:
        try:
            response = self.session.get(f"{self.rest_api_url}/data-sets/{dataset_id}")
            response.raise_for_status()
            response_dict = response.json()
            response_dict[Constant.ITEMID] = item_id
            return QlikDataset.parse_obj(response_dict)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch dataset with id {dataset_id}. Exception: {e}"
            )
        return None

    def get_user_name(self, user_id: str) -> Optional[str]:
        try:
            if user_id in self.users:
                # To avoid fetching same user details again
                return self.users[user_id]
            else:
                response = self.session.get(f"{self.rest_api_url}/users/{user_id}")
                response.raise_for_status()
                user_name = response.json()[Constant.NAME]
                self.users[user_id] = user_name
                return user_name
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch user with id {user_id}. Exception: {e}"
            )
        return None

    def _get_chart(
        self,
        websocket_connection: WebsocketConnection,
        chart_id: str,
        sheet_id: str,
    ) -> Optional[Chart]:
        try:
            websocket_connection.websocket_send_request(
                method="GetChild", params={"qId": chart_id}
            )
            response = websocket_connection.websocket_send_request(method="GetLayout")
            q_layout = response[Constant.QLAYOUT]
            if Constant.HYPERCUBE not in q_layout:
                logger.warning(
                    f"Chart with id {chart_id} of sheet {sheet_id} does not have hypercube. q_layout: {q_layout}"
                )
                return None
            return Chart.parse_obj(q_layout)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch chart {chart_id} of sheet {sheet_id}. Exception: {e}"
            )
        return None

    def _get_sheet(
        self,
        websocket_connection: WebsocketConnection,
        sheet_id: str,
    ) -> Optional[Sheet]:
        try:
            websocket_connection.websocket_send_request(
                method="GetObject", params={"qId": sheet_id}
            )
            response = websocket_connection.websocket_send_request(method="GetLayout")
            sheet_dict = response[Constant.QLAYOUT]
            if Constant.OWNERID not in sheet_dict[Constant.QMETA]:
                # That means sheet is private sheet
                return None
            sheet = Sheet.parse_obj(sheet_dict[Constant.QMETA])
            if Constant.QCHILDLIST not in sheet_dict:
                logger.warning(
                    f"Sheet {sheet.title} with id {sheet_id} does not have any charts. sheet_dict: {sheet_dict}"
                )
                return sheet
            for i, chart_dict in enumerate(
                sheet_dict[Constant.QCHILDLIST][Constant.QITEMS]
            ):
                chart = self._get_chart(
                    websocket_connection,
                    chart_dict[Constant.QINFO][Constant.QID],
                    sheet_id,
                )
                if chart:
                    if not chart.title:
                        chart.title = f"Object {i + 1} of Sheet '{sheet.title}'"
                    sheet.charts.append(chart)
                websocket_connection.handle.pop()
            return sheet
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch sheet with id {sheet_id}. Exception: {e}"
            )
        return None

    def _add_qri_of_tables(self, tables: List[QlikTable], app_id: str) -> None:
        table_qri_dict: Dict[str, str] = {}
        app_qri = quote(f"qri:app:sense://{app_id}", safe="")
        try:
            response = self.session.get(
                f"{self.rest_api_url}/lineage-graphs/nodes/{app_qri}/actions/expand?node={app_qri}&level=TABLE"
            )
            response.raise_for_status()
            for table_node_qri in response.json()[Constant.GRAPH][Constant.NODES]:
                table_node_qri = quote(table_node_qri, safe="")
                response = self.session.get(
                    f"{self.rest_api_url}/lineage-graphs/nodes/{app_qri}/actions/expand?node={table_node_qri}&level=FIELD"
                )
                response.raise_for_status()
                field_nodes_qris = list(
                    response.json()[Constant.GRAPH][Constant.NODES].keys()
                )
                for field_node_qri in field_nodes_qris:
                    response = self.session.post(
                        f"{self.rest_api_url}/lineage-graphs/nodes/{app_qri}/overview",
                        json=[field_node_qri],
                    )
                    response.raise_for_status()
                    # Some fields might not have lineage overview, in that case status code is 207
                    if response.status_code == 200:
                        for each_lineage in response.json()[Constant.RESOURCES][0][
                            Constant.LINEAGE
                        ]:
                            table_name = (
                                each_lineage[Constant.TABLELABEL]
                                .replace('"', "")
                                .split(".")[-1]
                            )
                            table_qri_dict[table_name] = each_lineage[Constant.TABLEQRI]
                        break
            for table in tables:
                if table.tableName in table_qri_dict:
                    table.tableQri = table_qri_dict[table.tableName]
        except Exception as e:
            self._log_http_error(
                message=f"Unable to add QRI for tables of app {app_id}. Exception: {e}"
            )

    def _get_app_used_tables(
        self, websocket_connection: WebsocketConnection, app_id: str
    ) -> List[QlikTable]:
        tables: List[QlikTable] = []
        try:
            response = websocket_connection.websocket_send_request(
                method="GetObject",
                params=["LoadModel"],
            )
            if not response[Constant.QRETURN][Constant.QTYPE]:
                return []
            response = websocket_connection.websocket_send_request(method="GetLayout")
            for table_dict in response[Constant.QLAYOUT][Constant.TABLES]:
                tables.append(QlikTable.parse_obj(table_dict))
            websocket_connection.handle.pop()
            self._add_qri_of_tables(tables, app_id)
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch tables used by app {app_id}. Exception: {e}"
            )
        return tables

    def _get_app_sheets(
        self, websocket_connection: WebsocketConnection, app_id: str
    ) -> List[Sheet]:
        sheets: List[Sheet] = []
        try:
            response = websocket_connection.websocket_send_request(
                method="GetObjects",
                params={
                    "qOptions": {
                        "qTypes": ["sheet"],
                    }
                },
            )
            for sheet_dict in response[Constant.QLIST]:
                sheet = self._get_sheet(
                    websocket_connection=websocket_connection,
                    sheet_id=sheet_dict[Constant.QINFO][Constant.QID],
                )
                if sheet:
                    sheets.append(sheet)
                websocket_connection.handle.pop()
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch sheets for app {app_id}. Exception: {e}"
            )
        return sheets

    def _get_app(self, app_id: str) -> Optional[App]:
        try:
            websocket_connection = WebsocketConnection(
                self.config.tenant_hostname, self.config.api_key, app_id
            )
            websocket_connection.websocket_send_request(
                method="OpenDoc",
                params={"qDocName": app_id},
            )
            response = websocket_connection.websocket_send_request(
                method="GetAppLayout"
            )
            app = App.parse_obj(response[Constant.QLAYOUT])
            app.sheets = self._get_app_sheets(websocket_connection, app_id)
            app.tables = self._get_app_used_tables(websocket_connection, app_id)
            websocket_connection.close_websocket()
            return app
        except Exception as e:
            self._log_http_error(
                message=f"Unable to fetch app with id {app_id}. Exception: {e}"
            )
        return None

    def get_items(self) -> List[Item]:
        items: List[Item] = []
        try:
            url = f"{self.rest_api_url}/items"
            while True:
                response = self.session.get(url)
                response.raise_for_status()
                response_dict = response.json()
                for item in response_dict[Constant.DATA]:
                    # spaceId none indicates item present in personal space
                    if not item.get(Constant.SPACEID):
                        item[Constant.SPACEID] = Constant.PERSONAL_SPACE_ID
                    if self.config.space_pattern.allowed(
                        self.spaces[item[Constant.SPACEID]]
                    ):
                        resource_type = item[Constant.RESOURCETYPE]
                        if resource_type == Constant.APP:
                            app = self._get_app(app_id=item[Constant.RESOURCEID])
                            if app:
                                items.append(app)
                        elif resource_type == Constant.DATASET:
                            dataset = self._get_dataset(
                                dataset_id=item[Constant.RESOURCEID],
                                item_id=item[Constant.ID],
                            )
                            if dataset:
                                items.append(dataset)
                if Constant.NEXT in response_dict[Constant.LINKS]:
                    url = response_dict[Constant.LINKS][Constant.NEXT][Constant.HREF]
                else:
                    break

        except Exception as e:
            self._log_http_error(message=f"Unable to fetch items. Exception: {e}")
        return items



import glob
import io
import logging
import os
import re
from datetime import datetime, timezone
from enum import Enum, auto
from io import BytesIO
from pathlib import PurePath
from typing import Any, Dict, Iterable, Iterator, List, Optional, Tuple, Type, Union
from urllib.parse import urlparse

from datahub.emitter.mce_builder import make_dataset_urn_with_platform_instance
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceCapability
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.aws.s3_boto_utils import get_s3_tags
from datahub.ingestion.source.aws.s3_util import (
    get_bucket_name,
    get_bucket_relative_path,
    strip_s3_prefix,
)
from datahub.ingestion.source.azure.abs_folder_utils import (
    get_abs_tags,
)
from datahub.ingestion.source.azure.abs_utils import (
    get_container_relative_path,
    strip_abs_prefix,
)
from datahub.ingestion.source.data_lake_common.data_lake_utils import ContainerWUCreator
from datahub.ingestion.source.excel.config import ExcelSourceConfig
from datahub.ingestion.source.excel.excel_file import ExcelFile, ExcelTable
from datahub.ingestion.source.excel.profiling import ExcelProfiler
from datahub.ingestion.source.excel.report import ExcelSourceReport
from datahub.ingestion.source.excel.util import gen_dataset_name
from datahub.ingestion.source.s3.source import BrowsePath
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import TimeStamp
from datahub.metadata.schema_classes import (
    BooleanTypeClass,
    ChangeTypeClass,
    DatasetPropertiesClass,
    DateTypeClass,
    GlobalTagsClass,
    NullTypeClass,
    NumberTypeClass,
    OtherSchemaClass,
    RecordTypeClass,
    SchemaFieldClass as SchemaField,
    SchemaFieldDataTypeClass as SchemaFieldDataType,
    SchemaMetadataClass as SchemaMetadata,
    StringTypeClass,
)
from datahub.utilities.perf_timer import PerfTimer

logger: logging.Logger = logging.getLogger(__name__)

field_type_mapping: Dict[str, Type] = {
    "int8": NumberTypeClass,
    "int16": NumberTypeClass,
    "int32": NumberTypeClass,
    "int64": NumberTypeClass,
    "uint8": NumberTypeClass,
    "uint16": NumberTypeClass,
    "uint32": NumberTypeClass,
    "uint64": NumberTypeClass,
    "Int8": NumberTypeClass,
    "Int16": NumberTypeClass,
    "Int32": NumberTypeClass,
    "Int64": NumberTypeClass,
    "UInt8": NumberTypeClass,
    "UInt16": NumberTypeClass,
    "UInt32": NumberTypeClass,
    "UInt64": NumberTypeClass,
    "intp": NumberTypeClass,
    "uintp": NumberTypeClass,
    "float16": NumberTypeClass,
    "float32": NumberTypeClass,
    "float64": NumberTypeClass,
    "float128": NumberTypeClass,
    "Float32": NumberTypeClass,
    "Float64": NumberTypeClass,
    "complex64": NumberTypeClass,
    "complex128": NumberTypeClass,
    "complex256": NumberTypeClass,
    "bool": BooleanTypeClass,
    "boolean": BooleanTypeClass,
    "object": StringTypeClass,
    "string": StringTypeClass,
    "datetime64": DateTypeClass,
    "datetime64[ns]": DateTypeClass,
    "datetime64[ns, tz]": DateTypeClass,
    "timedelta64": DateTypeClass,
    "timedelta64[ns]": DateTypeClass,
    "period": DateTypeClass,
    "period[D]": DateTypeClass,
    "period[M]": DateTypeClass,
    "period[Y]": DateTypeClass,
    "category": RecordTypeClass,
    "interval": RecordTypeClass,
    "sparse": RecordTypeClass,
    "NA": NullTypeClass,
}


ALLOWED_EXTENSIONS = [".xlsx", ".xlsm", ".xltx", ".xltm"]


class UriType(Enum):
    HTTP = auto()
    HTTPS = auto()
    LOCAL_FILE = auto()
    ABSOLUTE_PATH = auto()
    RELATIVE_PATH = auto()
    S3 = auto()
    S3A = auto()
    ABS = auto()
    UNKNOWN = auto()


@platform_name("Excel")
@config_class(ExcelSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.CONTAINERS, "Enabled by default")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@capability(SourceCapability.SCHEMA_METADATA, "Enabled by default")
@capability(
    SourceCapability.DELETION_DETECTION,
    "Optionally enabled via `stateful_ingestion.remove_stale_metadata`",
    supported=True,
)
class ExcelSource(StatefulIngestionSourceBase):
    config: ExcelSourceConfig
    report: ExcelSourceReport
    container_WU_creator: ContainerWUCreator
    platform: str = "excel"

    def __init__(self, ctx: PipelineContext, config: ExcelSourceConfig):
        super().__init__(config, ctx)
        self.ctx = ctx
        self.config = config
        self.report: ExcelSourceReport = ExcelSourceReport()

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "ExcelSource":
        config = ExcelSourceConfig.parse_obj(config_dict)
        return cls(ctx, config)

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    @staticmethod
    def uri_type(uri: str) -> Tuple[UriType, str]:
        if not uri or not isinstance(uri, str):
            return UriType.UNKNOWN, ""

        uri = uri.strip()
        parsed = urlparse(uri)
        scheme = parsed.scheme.lower()

        if scheme == "http":
            return UriType.HTTP, uri[7:]
        elif scheme == "https":
            if parsed.netloc and ".blob.core.windows.net" in parsed.netloc:
                return UriType.ABS, uri[8:]
            else:
                return UriType.HTTPS, uri[8:]
        elif scheme == "file":
            if uri.startswith("file:///"):
                return UriType.LOCAL_FILE, uri[7:]

        if scheme == "s3":
            return UriType.S3, uri[5:]
        elif scheme == "s3a":
            return UriType.S3A, uri[6:]

        if scheme:
            return UriType.UNKNOWN, uri[len(scheme) + 3 :]

        if os.path.isabs(uri):
            return UriType.ABSOLUTE_PATH, uri
        else:
            return UriType.RELATIVE_PATH, uri

    @staticmethod
    def is_excel_file(path: str) -> bool:
        _, ext = os.path.splitext(path)
        return ext.lower() in ALLOWED_EXTENSIONS

    @staticmethod
    def local_browser(path_spec: str) -> Iterable[BrowsePath]:
        matching_paths = glob.glob(path_spec, recursive=True)
        matching_files = [path for path in matching_paths if os.path.isfile(path)]

        for file in sorted(matching_files):
            full_path = PurePath(os.path.normpath(file)).as_posix()
            yield BrowsePath(
                file=full_path,
                timestamp=datetime.fromtimestamp(
                    os.path.getmtime(full_path), timezone.utc
                ),
                size=os.path.getsize(full_path),
                partitions=[],
            )

    def get_local_file(self, file_path: str) -> Union[BytesIO, None]:
        try:
            with open(file_path, "rb") as f:
                bytes_io = io.BytesIO(f.read())
                bytes_io.seek(0)
                return bytes_io
        except Exception as e:
            self.report.report_file_dropped(file_path)
            self.report.warning(
                message="Error reading local Excel file",
                context=f"Path={file_path}",
                exc=e,
            )
            return None

    @staticmethod
    def get_prefix(relative_path: str) -> str:
        index = re.search(r"[*|{]", relative_path)
        if index:
            return relative_path[: index.start()]
        else:
            return relative_path

    @staticmethod
    def create_s3_path(bucket_name: str, key: str) -> str:
        return f"s3://{bucket_name}/{key}"

    def create_abs_path(self, key: str) -> str:
        if self.config.azure_config:
            account_name = self.config.azure_config.account_name
            container_name = self.config.azure_config.container_name
            return (
                f"https://{account_name}.blob.core.windows.net/{container_name}/{key}"
            )
        return ""

    @staticmethod
    def strip_file_prefix(path: str) -> str:
        if path.startswith("/"):
            return path[1:]
        else:
            return path

    def s3_browser(self, path_spec: str) -> Iterable[BrowsePath]:
        if self.config.aws_config is None:
            raise ValueError("aws_config not set. Cannot browse s3")
        s3 = self.config.aws_config.get_s3_resource(self.config.verify_ssl)
        bucket_name = get_bucket_name(path_spec)
        logger.debug(f"Scanning bucket: {bucket_name}")
        bucket = s3.Bucket(bucket_name)
        prefix = self.get_prefix(get_bucket_relative_path(path_spec))
        logger.debug(f"Scanning objects with prefix:{prefix}")

        for obj in bucket.objects.filter(Prefix=prefix).page_size(1000):
            s3_path = self.create_s3_path(obj.bucket_name, obj.key)
            logger.debug(f"Path: {s3_path}")

            yield BrowsePath(
                file=s3_path,
                timestamp=obj.last_modified,
                size=obj.size,
                partitions=[],
                content_type=None,
            )

    def get_s3_file(self, path_spec: str) -> Union[BytesIO, None]:
        if self.config.aws_config is None:
            raise ValueError("aws_config not set. Cannot browse s3")
        s3 = self.config.aws_config.get_s3_resource(self.config.verify_ssl)
        bucket_name = get_bucket_name(path_spec)
        key = get_bucket_relative_path(path_spec)
        logger.debug(f"Getting file: {key} from bucket: {bucket_name}")
        try:
            obj = s3.Object(bucket_name, key)
            file_content = obj.get()["Body"].read()
            binary_stream = io.BytesIO(file_content)
            binary_stream.seek(0)
            return binary_stream
        except Exception as e:
            self.report.report_file_dropped(path_spec)
            self.report.warning(
                message="Error reading Excel file from S3",
                context=f"Path={path_spec}",
                exc=e,
            )
            return None

    def process_s3_tags(
        self, path_spec: str, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        bucket_name = get_bucket_name(path_spec)
        key = get_bucket_relative_path(path_spec)

        s3_tags = get_s3_tags(
            bucket_name,
            key,
            dataset_urn,
            self.config.aws_config,
            self.ctx,
            self.config.use_s3_bucket_tags,
            self.config.use_s3_object_tags,
            self.config.verify_ssl,
        )

        if s3_tags:
            yield from self.process_global_tags(s3_tags, dataset_urn)

    def abs_browser(self, path_spec: str) -> Iterable[BrowsePath]:
        if self.config.azure_config is None:
            raise ValueError("azure_config not set. Cannot browse Azure Blob Storage")
        abs_blob_service_client = self.config.azure_config.get_blob_service_client()
        container_client = abs_blob_service_client.get_container_client(
            self.config.azure_config.container_name
        )

        container_name = self.config.azure_config.container_name
        logger.debug(f"Scanning container: {container_name}")

        prefix = self.get_prefix(get_container_relative_path(path_spec))
        logger.debug(f"Scanning objects with prefix: {prefix}")

        for obj in container_client.list_blobs(
            name_starts_with=f"{prefix}", results_per_page=1000
        ):
            abs_path = self.create_abs_path(obj.name)
            logger.debug(f"Path: {abs_path}")

            yield BrowsePath(
                file=abs_path,
                timestamp=obj.last_modified,
                size=obj.size,
                partitions=[],
                content_type=None,
            )

    def get_abs_file(self, path_spec: str) -> Union[BytesIO, None]:
        if self.config.azure_config is None:
            raise ValueError("azure_config not set. Cannot browse Azure Blob Storage")
        abs_blob_service_client = self.config.azure_config.get_blob_service_client()
        container_client = abs_blob_service_client.get_container_client(
            self.config.azure_config.container_name
        )

        container_name = self.config.azure_config.container_name
        blob_path = get_container_relative_path(path_spec)
        logger.debug(f"Getting file: {blob_path} from container: {container_name}")

        try:
            blob_client = container_client.get_blob_client(blob_path)
            download_stream = blob_client.download_blob()
            file_content = download_stream.readall()
            binary_stream = io.BytesIO(file_content)
            binary_stream.seek(0)
            return binary_stream
        except Exception as e:
            self.report.report_file_dropped(path_spec)
            self.report.warning(
                message="Error reading Excel file from Azure Blob Storage",
                context=f"Path={path_spec}",
                exc=e,
            )
            return None

    def process_abs_tags(
        self, path_spec: str, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        if (
            self.config.azure_config
            and self.config.azure_config.container_name is not None
        ):
            container_name = self.config.azure_config.container_name
            blob_path = get_container_relative_path(path_spec)

            abs_tags = get_abs_tags(
                container_name,
                blob_path,
                dataset_urn,
                self.config.azure_config,
                self.ctx,
                self.config.use_abs_blob_tags,
            )

            if abs_tags:
                yield from self.process_global_tags(abs_tags, dataset_urn)

    @staticmethod
    def get_field_type(field_type: str) -> SchemaFieldDataType:
        type_class = field_type_mapping.get(field_type, NullTypeClass)
        return SchemaFieldDataType(type=type_class())

    def construct_schema_field(self, f_name: str, f_type: str) -> SchemaField:
        logger.debug(f"Field: {f_name} Type: {f_type}")
        return SchemaField(
            fieldPath=f_name,
            nativeDataType=f_type,
            type=self.get_field_type(f_type),
            description=None,
            nullable=False,
            recursive=False,
        )

    def construct_schema_metadata(
        self,
        name: str,
        dataset: ExcelTable,
    ) -> SchemaMetadata:
        canonical_schema: List[SchemaField] = []

        # Get data types for each column
        data_types = dataset.df.dtypes.to_dict()

        # Convert numpy types to string representation for better readability
        data_types = {col: str(dtype) for col, dtype in data_types.items()}

        for f_name, f_type in data_types.items():
            canonical_schema.append(self.construct_schema_field(f_name, f_type))

        return SchemaMetadata(
            schemaName=name,
            platform=f"urn:li:dataPlatform:{self.platform}",
            version=0,
            hash="",
            platformSchema=OtherSchemaClass(rawSchema=""),
            fields=canonical_schema,
        )

    @staticmethod
    def get_dataset_attributes(metadata: Dict[str, Any]) -> dict:
        result = {}
        for key, value in metadata.items():
            result[key] = str(value)
        return result

    @staticmethod
    def process_global_tags(
        global_tags: GlobalTagsClass, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        yield MetadataChangeProposalWrapper(
            entityType="dataset",
            entityUrn=dataset_urn,
            aspect=global_tags,
            changeType=ChangeTypeClass.UPSERT,
        ).as_workunit()

    def process_dataset(
        self,
        relative_path: str,
        full_path: str,
        filename: str,
        table: ExcelTable,
        source_type: UriType,
    ) -> Iterable[MetadataWorkUnit]:
        self.report.report_worksheet_processed()
        dataset_name = gen_dataset_name(
            relative_path, table.sheet_name, self.config.convert_urns_to_lowercase
        )
        dataset_urn = make_dataset_urn_with_platform_instance(
            platform=self.platform,
            name=dataset_name,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
        )

        attributes = self.get_dataset_attributes(table.metadata)
        created: Optional[datetime] = table.metadata.get("created")
        modified: Optional[datetime] = table.metadata.get("modified")
        dataset_properties = DatasetPropertiesClass(
            tags=[],
            customProperties=attributes,
            created=(
                TimeStamp(time=int(created.timestamp() * 1000)) if created else None
            ),
            lastModified=(
                TimeStamp(time=int(modified.timestamp() * 1000)) if modified else None
            ),
        )

        schema_metadata = self.construct_schema_metadata(
            name=dataset_name,
            dataset=table,
        )

        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=schema_metadata,
        ).as_workunit()

        yield MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=dataset_properties,
        ).as_workunit()

        yield from self.container_WU_creator.create_container_hierarchy(
            relative_path, dataset_urn
        )

        if source_type == UriType.S3 and (
            self.config.use_s3_bucket_tags or self.config.use_s3_object_tags
        ):
            yield from self.process_s3_tags(full_path, dataset_urn)
        elif source_type == UriType.ABS and self.config.use_abs_blob_tags:
            yield from self.process_abs_tags(full_path, dataset_urn)

        if self.config.is_profiling_enabled():
            profiler = ExcelProfiler(
                self.config,
                self.report,
                table.df,
                filename,
                table.sheet_name,
                dataset_urn,
                relative_path,
            )
            yield from profiler.get_workunits()

    def process_file(
        self,
        file_content: BytesIO,
        relative_path: str,
        full_path: str,
        filename: str,
        source_type: UriType,
    ) -> Iterable[MetadataWorkUnit]:
        self.report.report_file_processed()
        xls = ExcelFile(filename, file_content, self.report)
        result = xls.load_workbook()

        if result:
            for table in xls.get_tables(active_only=self.config.active_sheet_only):
                self.report.report_worksheet_scanned()
                dataset_name = gen_dataset_name(
                    relative_path,
                    table.sheet_name,
                    self.config.convert_urns_to_lowercase,
                )
                if not self.config.worksheet_pattern.allowed(dataset_name):
                    self.report.report_dropped(dataset_name)
                    continue
                yield from self.process_dataset(
                    relative_path, full_path, filename, table, source_type
                )

    def check_file_is_valid(self, filename: str) -> bool:
        self.report.report_file_scanned()
        if not self.config.path_pattern.allowed(filename):
            self.report.report_dropped(filename)
            return False
        elif not self.is_excel_file(filename):
            logger.debug(f"File is not an Excel workbook: {filename}")
            return False
        return True

    def retrieve_file_data(
        self, uri_type: UriType, path: str, path_spec: str
    ) -> Iterator[Tuple[BytesIO, str, str, str]]:
        if (
            uri_type == UriType.LOCAL_FILE
            or uri_type == UriType.ABSOLUTE_PATH
            or uri_type == UriType.RELATIVE_PATH
        ):
            logger.debug(f"Searching local path: {path}")
            for browse_path in self.local_browser(path):
                if self.check_file_is_valid(browse_path.file):
                    basename = os.path.basename(browse_path.file)
                    file_path = self.strip_file_prefix(browse_path.file)
                    filename = os.path.splitext(basename)[0]

                    logger.debug(f"Processing {filename}")
                    with self.report.local_file_get_timer:
                        file_data = self.get_local_file(browse_path.file)

                    if file_data is not None:
                        yield file_data, file_path, browse_path.file, filename

        elif uri_type == UriType.S3:
            logger.debug(f"Searching S3 path: {path}")
            for browse_path in self.s3_browser(path_spec):
                if self.check_file_is_valid(browse_path.file):
                    uri_path = strip_s3_prefix(browse_path.file)
                    basename = os.path.basename(uri_path)
                    filename = os.path.splitext(basename)[0]

                    logger.debug(f"Processing {browse_path.file}")
                    with self.report.s3_file_get_timer:
                        file_data = self.get_s3_file(browse_path.file)

                    if file_data is not None:
                        yield file_data, uri_path, browse_path.file, filename

        elif uri_type == UriType.ABS:
            logger.debug(f"Searching Azure Blob Storage path: {path}")
            for browse_path in self.abs_browser(path_spec):
                if self.check_file_is_valid(browse_path.file):
                    uri_path = strip_abs_prefix(browse_path.file)
                    basename = os.path.basename(uri_path)
                    filename = os.path.splitext(basename)[0]

                    logger.debug(f"Processing {browse_path.file}")
                    with self.report.abs_file_get_timer:
                        file_data = self.get_abs_file(browse_path.file)

                    if file_data is not None:
                        yield file_data, uri_path, browse_path.file, filename

        else:
            self.report.report_file_dropped(path_spec)
            self.report.warning(
                message="Unsupported URI Type",
                context=f"Type={uri_type.name},URI={path_spec}",
            )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        self.container_WU_creator = ContainerWUCreator(
            self.platform,
            self.config.platform_instance,
            self.config.env,
        )

        with PerfTimer() as timer:
            for path_spec in self.config.path_list:
                logger.debug(f"Processing path: {path_spec}")
                uri_type, path = self.uri_type(path_spec)
                logger.debug(f"URI Type: {uri_type} Path: {path}")

                for (
                    file_data,
                    relative_path,
                    full_path,
                    filename,
                ) in self.retrieve_file_data(uri_type, path, path_spec):
                    yield from self.process_file(
                        file_data, relative_path, full_path, filename, uri_type
                    )

        time_taken = timer.elapsed_seconds()
        logger.info(f"Finished ingestion in {time_taken:.3f} seconds")

    def get_report(self):
        return self.report



import datetime
import itertools
import logging
import os
import re
from contextlib import contextmanager
from dataclasses import dataclass, field as dataclasses_field
from enum import Enum
from functools import lru_cache
from typing import (
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    cast,
)

from looker_sdk.error import SDKError
from looker_sdk.rtl.serialize import DeserializeError
from looker_sdk.sdk.api40.models import (
    LookmlModelExplore,
    LookmlModelExploreField,
    User,
    WriteQuery,
)
from pydantic import validator

import datahub.emitter.mce_builder as builder
from datahub.api.entities.platformresource.platform_resource import (
    PlatformResource,
    PlatformResourceKey,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import ContainerKey
from datahub.ingestion.api.report import Report
from datahub.ingestion.api.source import SourceReport
from datahub.ingestion.source.common.subtypes import DatasetSubTypes
from datahub.ingestion.source.looker.looker_config import (
    LookerCommonConfig,
    LookerConnectionDefinition,
    LookerDashboardSourceConfig,
    NamingPatternMapping,
    ViewNamingPatternMapping,
)
from datahub.ingestion.source.looker.looker_constant import IMPORTED_PROJECTS
from datahub.ingestion.source.looker.looker_dataclasses import ProjectInclude
from datahub.ingestion.source.looker.looker_file_loader import LookerViewFileLoader
from datahub.ingestion.source.looker.looker_lib_wrapper import LookerAPI
from datahub.ingestion.source.looker.lookml_config import (
    BASE_PROJECT_NAME,
    LookMLSourceReport,
)
from datahub.ingestion.source.looker.str_functions import remove_suffix
from datahub.ingestion.source.sql.sql_types import (
    POSTGRES_TYPES_MAP,
    SNOWFLAKE_TYPES_MAP,
    resolve_postgres_modified_type,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import AuditStamp
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    DatasetLineageTypeClass,
    FineGrainedLineageDownstreamType,
    FineGrainedLineageUpstreamType,
    UpstreamClass,
    UpstreamLineage,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayTypeClass,
    BooleanTypeClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    OtherSchema,
    SchemaField,
    SchemaFieldDataType,
    SchemaMetadata,
    StringTypeClass,
    TimeTypeClass,
    UnionTypeClass,
)
from datahub.metadata.schema_classes import (
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    EmbedClass,
    EnumTypeClass,
    FineGrainedLineageClass,
    GlobalTagsClass,
    SchemaMetadataClass,
    TagAssociationClass,
    TagPropertiesClass,
    TagSnapshotClass,
)
from datahub.metadata.urns import TagUrn
from datahub.sdk.dataset import Dataset
from datahub.sql_parsing.sqlglot_lineage import ColumnRef
from datahub.utilities.lossy_collections import LossyList, LossySet
from datahub.utilities.url_util import remove_port_from_url

CORPUSER_DATAHUB = "urn:li:corpuser:datahub"
LOOKER = "looker"
logger = logging.getLogger(__name__)


@dataclass
class LookerFolder:
    id: str
    name: str
    parent_id: Optional[str]


class LookMLProjectKey(ContainerKey):
    project_name: str


class LookMLModelKey(ContainerKey):
    model_name: str


class LookerFolderKey(ContainerKey):
    folder_id: str


def deduplicate_fields(fields: List["ViewField"]) -> List["ViewField"]:
    # Remove duplicates filed from self.fields
    # Logic is: If more than a field has same ViewField.name then keep only one filed where ViewField.field_type
    # is DIMENSION_GROUP.
    # Looker Constraint:
    #   - Any field declared as dimension or measure can be redefined as dimension_group.
    #   - Any field declared in dimension can't be redefined in measure and vice-versa.

    dimension_group_field_names: List[str] = [
        field.name
        for field in fields
        if field.field_type == ViewFieldType.DIMENSION_GROUP
    ]

    new_fields: List[ViewField] = []

    for field in fields:
        if (
            field.name in dimension_group_field_names
            and field.field_type != ViewFieldType.DIMENSION_GROUP
        ):
            continue

        new_fields.append(field)

    return new_fields


def find_view_from_resolved_includes(
    connection: Optional[LookerConnectionDefinition],
    resolved_includes: List["ProjectInclude"],
    looker_viewfile_loader: LookerViewFileLoader,
    target_view_name: str,
    reporter: LookMLSourceReport,
) -> Optional[Tuple["ProjectInclude", dict]]:
    # It could live in one of the included files. We do not know which file the base view
    # lives in, so we try them all!
    for include in resolved_includes:
        included_looker_viewfile = looker_viewfile_loader.load_viewfile(
            include.include,
            include.project,
            connection,
            reporter,
        )
        if not included_looker_viewfile:
            continue
        for raw_view in included_looker_viewfile.views:
            raw_view_name = raw_view["name"]
            # Make sure to skip loading view we are currently trying to resolve
            if raw_view_name == target_view_name:
                return include, raw_view

    return None


@dataclass
class LookerViewId:
    project_name: str
    model_name: str
    view_name: str
    file_path: str

    def get_mapping(self, config: LookerCommonConfig) -> ViewNamingPatternMapping:
        return ViewNamingPatternMapping(
            platform=config.platform_name,
            env=config.env.lower(),
            project=self.project_name,
            model=self.model_name,
            name=self.view_name,
            file_path=remove_suffix(self.file_path, ".view.lkml"),
            folder_path=os.path.dirname(self.file_path),
        )

    @validator("view_name")
    def remove_quotes(cls, v):
        # Sanitize the name.
        v = v.replace('"', "").replace("`", "")
        return v

    def preprocess_file_path(self, file_path: str) -> str:
        new_file_path: str = str(file_path)

        str_to_remove: List[str] = [
            "\\.view\\.lkml$",  # escape the . using \
        ]

        for pattern in str_to_remove:
            new_file_path = re.sub(pattern, "", new_file_path)

        str_to_replace: Dict[str, str] = {
            f"^imported_projects/{re.escape(self.project_name)}/": "",  # escape any special regex character present in project-name
            "/": ".",  # / is not urn friendly
        }

        for pattern in str_to_replace:
            new_file_path = re.sub(pattern, str_to_replace[pattern], new_file_path)

        logger.debug(f"Original file path {file_path}")
        logger.debug(f"After preprocessing file path {new_file_path}")

        return new_file_path

    def get_urn(self, config: LookerCommonConfig) -> str:
        n_mapping: ViewNamingPatternMapping = self.get_mapping(config)

        n_mapping.file_path = self.preprocess_file_path(n_mapping.file_path)

        dataset_name = config.view_naming_pattern.replace_variables(n_mapping)

        generated_urn = builder.make_dataset_urn_with_platform_instance(
            platform=config.platform_name,
            name=dataset_name,
            platform_instance=config.platform_instance,
            env=config.env,
        )

        logger.debug(
            f"LookerViewId.get_urn for view '{self.view_name}': project='{self.project_name}', model='{self.model_name}', file_path='{self.file_path}', dataset_name='{dataset_name}', generated_urn='{generated_urn}'"
        )

        return generated_urn

    def get_view_dataset_name(self, config: LookerCommonConfig) -> str:
        n_mapping: ViewNamingPatternMapping = self.get_mapping(config)
        n_mapping.file_path = self.preprocess_file_path(n_mapping.file_path)
        return config.view_naming_pattern.replace_variables(n_mapping)

    def get_browse_path(self, config: LookerCommonConfig) -> str:
        browse_path = config.view_browse_pattern.replace_variables(
            self.get_mapping(config)
        )
        return browse_path

    def get_browse_path_v2(self, config: LookerCommonConfig) -> BrowsePathsV2Class:
        project_key = gen_project_key(config, self.project_name)
        view_path = (
            remove_suffix(self.file_path, ".view.lkml")
            if "{file_path}" in config.view_browse_pattern.pattern
            else os.path.dirname(self.file_path)
        )
        if view_path:
            path_entries = [
                BrowsePathEntryClass(id=path) for path in view_path.split("/")
            ]
        else:
            path_entries = []
        return BrowsePathsV2Class(
            path=[
                BrowsePathEntryClass(id="Develop"),
                BrowsePathEntryClass(id=project_key.as_urn(), urn=project_key.as_urn()),
                *path_entries,
            ],
        )

    def get_view_dataset_parent_container(
        self, config: LookerCommonConfig
    ) -> List[str]:
        project_key = gen_project_key(config, self.project_name)
        view_path = (
            remove_suffix(self.file_path, ".view.lkml")
            if "{file_path}" in config.view_browse_pattern.pattern
            else os.path.dirname(self.file_path)
        )
        path_entries = view_path.split("/") if view_path else []
        return [
            "Develop",
            project_key.as_urn(),
            *path_entries,
        ]


class ViewFieldType(Enum):
    DIMENSION = "Dimension"
    DIMENSION_GROUP = "Dimension Group"
    MEASURE = "Measure"
    UNKNOWN = "Unknown"


class ViewFieldValue(Enum):
    NOT_AVAILABLE = "NotAvailable"


@dataclass
class ViewField:
    name: str
    label: Optional[str]
    type: str
    description: str
    field_type: ViewFieldType
    project_name: Optional[str] = None
    view_name: Optional[str] = None
    is_primary_key: bool = False
    tags: List[str] = dataclasses_field(default_factory=list)
    group_label: Optional[str] = None

    # It is the list of ColumnRef for derived view defined using SQL otherwise simple column name
    upstream_fields: Union[List[ColumnRef]] = dataclasses_field(default_factory=list)

    @classmethod
    def view_fields_from_dict(
        cls,
        field_dict: Dict,
        upstream_column_ref: List[ColumnRef],
        type_cls: ViewFieldType,
        populate_sql_logic_in_descriptions: bool,
    ) -> "ViewField":
        is_primary_key = field_dict.get("primary_key", "no") == "yes"

        name = field_dict["name"]

        native_type = field_dict.get("type", "string")

        default_description = (
            f"sql:{field_dict['sql']}"
            if "sql" in field_dict and populate_sql_logic_in_descriptions
            else ""
        )

        description = field_dict.get("description", default_description)

        label = field_dict.get("label", "")
        group_label = field_dict.get("group_label")

        return ViewField(
            name=name,
            type=native_type,
            label=label,
            description=description,
            is_primary_key=is_primary_key,
            field_type=type_cls,
            tags=field_dict.get("tags") or [],
            upstream_fields=upstream_column_ref,
            group_label=group_label,
        )


@dataclass
class ExploreUpstreamViewField:
    explore: LookmlModelExplore
    field: LookmlModelExploreField

    def _form_field_name(
        self,
        view_project_map: Dict[str, str],
        explore_project_name: str,
        model_name: str,
        upstream_views_file_path: Dict[str, Optional[str]],
        config: LookerCommonConfig,
        remove_variant: bool = False,
    ) -> Optional[ColumnRef]:
        assert self.field.name is not None

        if len(self.field.name.split(".")) != 2:
            return None  # Inconsistent info received

        view_name: Optional[str] = self.explore.name
        if self.field.original_view is not None:
            view_name = self.field.original_view

        field_name = self.field.name.split(".")[1]

        if remove_variant and self.field.field_group_variant is not None:
            # remove variant at the end. +1 for "_"
            field_name = field_name[
                : -(len(self.field.field_group_variant.lower()) + 1)
            ]

        # Validate that field_name is not empty to prevent invalid schema field URNs
        if not field_name or not field_name.strip():
            logger.warning(
                f"Empty field name detected for field '{self.field.name}' in explore '{self.explore.name}'. "
                f"Skipping field to prevent invalid schema field URN generation."
            )
            return None

        assert view_name  # for lint false positive

        project_include: ProjectInclude = ProjectInclude(
            project=view_project_map.get(view_name, BASE_PROJECT_NAME),
            include=view_name,
        )

        file_path: Optional[str] = (
            upstream_views_file_path.get(view_name)
            if upstream_views_file_path.get(view_name)
            else ViewFieldValue.NOT_AVAILABLE.value
        )

        assert file_path

        view_urn = LookerViewId(
            project_name=(
                project_include.project
                if project_include.project != BASE_PROJECT_NAME
                else explore_project_name
            ),
            model_name=model_name,
            view_name=project_include.include,
            file_path=file_path,
        ).get_urn(config)

        return ColumnRef(
            table=view_urn,
            column=field_name,
        )

    def upstream(
        self,
        view_project_map: Dict[str, str],
        explore_project_name: str,
        model_name: str,
        upstream_views_file_path: Dict[str, Optional[str]],
        config: LookerCommonConfig,
    ) -> Optional[ColumnRef]:
        assert self.field.name is not None

        if self.field.dimension_group is None or self.field.field_group_variant is None:
            return self._form_field_name(
                view_project_map,
                explore_project_name,
                model_name,
                upstream_views_file_path,
                config,
            )

        if self.field.type is None or not self.field.type.startswith("date_"):
            return self._form_field_name(
                view_project_map,
                explore_project_name,
                model_name,
                upstream_views_file_path,
                config,
            )  # for Dimensional Group the type is always start with date_[time|date]

        if not self.field.name.endswith(f"_{self.field.field_group_variant.lower()}"):
            return self._form_field_name(
                view_project_map,
                explore_project_name,
                model_name,
                upstream_views_file_path,
                config,
            )  # if the explore field is generated because of  Dimensional Group in View
            # then the field_name should ends with field_group_variant

        return self._form_field_name(
            view_project_map,
            explore_project_name,
            model_name,
            upstream_views_file_path,
            config,
            remove_variant=True,
        )


def create_view_project_map(
    view_fields: List[ViewField],
    explore_primary_view: Optional[str] = None,
    explore_project_name: Optional[str] = None,
) -> Dict[str, str]:
    """
    Each view in a model has unique name.
    Use this function in scope of a model.

    Args:
        view_fields: List of ViewField objects
        explore_primary_view: The primary view name of the explore (explore.view_name)
        explore_project_name: The project name of the explore (explore.project_name)
    """
    view_project_map: Dict[str, str] = {}
    for view_field in view_fields:
        if view_field.view_name is not None and view_field.project_name is not None:
            # Override field-level project assignment for the primary view when different
            if (
                view_field.view_name == explore_primary_view
                and explore_project_name is not None
                and explore_project_name != view_field.project_name
            ):
                logger.debug(
                    f"Overriding project assignment for primary view '{view_field.view_name}': "
                    f"field-level project '{view_field.project_name}' → explore-level project '{explore_project_name}'"
                )
                view_project_map[view_field.view_name] = explore_project_name
            else:
                view_project_map[view_field.view_name] = view_field.project_name

    return view_project_map


def get_view_file_path(
    lkml_fields: List[LookmlModelExploreField], view_name: str
) -> Optional[str]:
    """
    Search for the view file path on field, if found then return the file path
    """
    logger.debug("Entered")

    for field in lkml_fields:
        if (
            LookerUtil.extract_view_name_from_lookml_model_explore_field(field)
            == view_name
        ):
            # This path is relative to git clone directory
            logger.debug(f"Found view({view_name}) file-path {field.source_file}")
            return field.source_file

    logger.debug(f"Failed to find view({view_name}) file-path")

    return None


def create_upstream_views_file_path_map(
    view_names: Set[str], lkml_fields: List[LookmlModelExploreField]
) -> Dict[str, Optional[str]]:
    """
    Create a map of view-name v/s view file path, so that later we can fetch view's file path via view-name
    """

    upstream_views_file_path: Dict[str, Optional[str]] = {}

    for view_name in view_names:
        file_path: Optional[str] = get_view_file_path(
            lkml_fields=lkml_fields, view_name=view_name
        )

        upstream_views_file_path[view_name] = file_path

    return upstream_views_file_path


def explore_field_set_to_lkml_fields(
    explore: LookmlModelExplore,
) -> List[LookmlModelExploreField]:
    """
    explore.fields has three variables i.e. dimensions, measures, parameters of same type i.e. LookmlModelExploreField.
    This method creating a list by adding all field instance to lkml_fields
    """
    lkml_fields: List[LookmlModelExploreField] = []

    if explore.fields is None:
        logger.debug(f"Explore({explore.name}) doesn't have any field")
        return lkml_fields

    def empty_list(
        fields: Optional[Sequence[LookmlModelExploreField]],
    ) -> List[LookmlModelExploreField]:
        return list(fields) if fields is not None else []

    lkml_fields.extend(empty_list(explore.fields.dimensions))
    lkml_fields.extend(empty_list(explore.fields.measures))
    lkml_fields.extend(empty_list(explore.fields.parameters))

    return lkml_fields


class LookerUtil:
    field_type_mapping = {
        **POSTGRES_TYPES_MAP,
        **SNOWFLAKE_TYPES_MAP,
        "date": DateTypeClass,
        "date_date": DateTypeClass,
        "date_day_of_month": NumberTypeClass,
        "date_day_of_week": EnumTypeClass,
        "date_day_of_week_index": EnumTypeClass,
        "date_fiscal_month_num": NumberTypeClass,
        "date_fiscal_quarter": DateTypeClass,
        "date_fiscal_quarter_of_year": EnumTypeClass,
        "date_hour": TimeTypeClass,
        "date_hour_of_day": NumberTypeClass,
        "date_month": DateTypeClass,
        "date_month_num": NumberTypeClass,
        "date_month_name": EnumTypeClass,
        "date_quarter": DateTypeClass,
        "date_quarter_of_year": EnumTypeClass,
        "date_time": TimeTypeClass,
        "date_time_of_day": TimeTypeClass,
        "date_microsecond": TimeTypeClass,
        "date_millisecond": TimeTypeClass,
        "date_minute": TimeTypeClass,
        "date_raw": TimeTypeClass,
        "date_second": TimeTypeClass,
        "date_week": TimeTypeClass,
        "date_year": DateTypeClass,
        "date_day_of_year": NumberTypeClass,
        "date_week_of_year": NumberTypeClass,
        "date_fiscal_year": DateTypeClass,
        "duration_day": StringTypeClass,
        "duration_hour": StringTypeClass,
        "duration_minute": StringTypeClass,
        "duration_month": StringTypeClass,
        "duration_quarter": StringTypeClass,
        "duration_second": StringTypeClass,
        "duration_week": StringTypeClass,
        "duration_year": StringTypeClass,
        "distance": NumberTypeClass,
        "duration": NumberTypeClass,
        "location": UnionTypeClass,
        "number": NumberTypeClass,
        "string": StringTypeClass,
        "tier": EnumTypeClass,
        "time": TimeTypeClass,
        "unquoted": StringTypeClass,
        "yesno": BooleanTypeClass,
        "zipcode": EnumTypeClass,
        "int": NumberTypeClass,
        "average": NumberTypeClass,
        "average_distinct": NumberTypeClass,
        "count": NumberTypeClass,
        "count_distinct": NumberTypeClass,
        "list": ArrayTypeClass,
        "max": NumberTypeClass,
        "median": NumberTypeClass,
        "median_distinct": NumberTypeClass,
        "min": NumberTypeClass,
        "percent_of_previous": NumberTypeClass,
        "percent_of_total": NumberTypeClass,
        "percentile": NumberTypeClass,
        "percentile_distinct": NumberTypeClass,
        "running_total": NumberTypeClass,
        "sum": NumberTypeClass,
        "sum_distinct": NumberTypeClass,
        "unknown": NullTypeClass,
    }

    @staticmethod
    def _extract_view_from_field(field: str) -> str:
        assert field.count(".") == 1, (
            f"Error: A field must be prefixed by a view name, field is: {field}"
        )
        return field.split(".")[0]

    @staticmethod
    def extract_view_name_from_lookml_model_explore_field(
        field: LookmlModelExploreField,
    ) -> Optional[str]:
        """
        View name is either present in original_view or view property
        """
        if field.original_view is not None:
            return field.original_view

        return field.view

    @staticmethod
    def extract_project_name_from_source_file(
        source_file: Optional[str],
    ) -> Optional[str]:
        """
        source_file is a key inside explore.fields. This key point to relative path of included views.
        if view is included from another project then source_file is starts with "imported_projects".
        Example: imported_projects/datahub-demo/views/datahub-demo/datasets/faa_flights.view.lkml
        """
        if source_file is None:
            return None

        if source_file.startswith(IMPORTED_PROJECTS):
            tokens: List[str] = source_file.split("/")
            if len(tokens) >= 2:
                return tokens[1]  # second index is project-name

        return None

    @staticmethod
    def get_field_type(native_type: str) -> SchemaFieldDataType:
        type_class = LookerUtil.field_type_mapping.get(native_type)

        if type_class is None:
            # attempt Postgres modified type
            type_class = resolve_postgres_modified_type(native_type)

        # if still not found, log and continue
        if type_class is None:
            logger.debug(
                f"The type '{native_type}' is not recognized for field type, setting as NullTypeClass.",
            )
            type_class = NullTypeClass

        return SchemaFieldDataType(type=type_class())

    @staticmethod
    def _get_schema(
        platform_name: str,
        schema_name: str,
        view_fields: List[ViewField],
        reporter: SourceReport,
    ) -> Optional[SchemaMetadataClass]:
        if not view_fields:
            return None
        fields, primary_keys = LookerUtil._get_fields_and_primary_keys(
            view_fields=view_fields, reporter=reporter
        )
        schema_metadata = SchemaMetadata(
            schemaName=schema_name,
            platform=f"urn:li:dataPlatform:{platform_name}",
            version=0,
            fields=fields,
            primaryKeys=primary_keys,
            hash="",
            platformSchema=OtherSchema(rawSchema=""),
        )
        return schema_metadata

    DIMENSION_TAG_URN = "urn:li:tag:Dimension"
    TEMPORAL_TAG_URN = "urn:li:tag:Temporal"
    MEASURE_TAG_URN = "urn:li:tag:Measure"

    type_to_tag_map: Dict[ViewFieldType, List[str]] = {
        ViewFieldType.DIMENSION: [DIMENSION_TAG_URN],
        ViewFieldType.DIMENSION_GROUP: [
            DIMENSION_TAG_URN,
            TEMPORAL_TAG_URN,
        ],
        ViewFieldType.MEASURE: [MEASURE_TAG_URN],
        ViewFieldType.UNKNOWN: [],
    }

    tag_definitions: Dict[str, TagPropertiesClass] = {
        DIMENSION_TAG_URN: TagPropertiesClass(
            name=DIMENSION_TAG_URN.split(":")[-1],
            description="A tag that is applied to all dimension fields.",
        ),
        TEMPORAL_TAG_URN: TagPropertiesClass(
            name=TEMPORAL_TAG_URN.split(":")[-1],
            description="A tag that is applied to all time-based (temporal) fields such as timestamps or durations.",
        ),
        MEASURE_TAG_URN: TagPropertiesClass(
            name=MEASURE_TAG_URN.split(":")[-1],
            description="A tag that is applied to all measures (metrics). Measures are typically the columns that you aggregate on",
        ),
    }

    # Add a pattern-based regex for checking if a tag is a group_label tag
    GROUP_LABEL_TAG_PATTERN = re.compile(r"^looker\:group_label\:(.+)$")

    @staticmethod
    def _get_tag_mce_for_urn(tag_urn: str) -> MetadataChangeEvent:
        # Check if this is a group_label tag
        tag_name = tag_urn[len("urn:li:tag:") :]
        match = LookerUtil.GROUP_LABEL_TAG_PATTERN.match(tag_name)

        if match:
            # This is a group_label tag, create tag definition on the fly
            group_label_value = match.group(1)
            tag_properties = TagPropertiesClass(
                name=f"looker:group_label:{group_label_value}",
                description=f"Fields with Looker group label: {group_label_value}",
            )

            return MetadataChangeEvent(
                proposedSnapshot=TagSnapshotClass(urn=tag_urn, aspects=[tag_properties])
            )
        elif tag_urn in LookerUtil.tag_definitions:
            # This is a predefined tag
            return MetadataChangeEvent(
                proposedSnapshot=TagSnapshotClass(
                    urn=tag_urn, aspects=[LookerUtil.tag_definitions[tag_urn]]
                )
            )
        else:
            # Should not happen, but handle gracefully
            logger.warning(f"No tag definition found for tag URN: {tag_urn}")
            return MetadataChangeEvent(
                proposedSnapshot=TagSnapshotClass(
                    urn=tag_urn,
                    aspects=[
                        TagPropertiesClass(
                            name=tag_name,
                            description=f"Tag: {tag_name}",
                        )
                    ],
                )
            )

    @staticmethod
    def _get_tags_from_field_type(
        field: ViewField, reporter: SourceReport
    ) -> Optional[GlobalTagsClass]:
        schema_field_tags: List[TagAssociationClass] = [
            TagAssociationClass(tag=builder.make_tag_urn(tag_name))
            for tag_name in field.tags
        ]

        if field.field_type in LookerUtil.type_to_tag_map:
            schema_field_tags.extend(
                [
                    TagAssociationClass(tag=tag_name)
                    for tag_name in LookerUtil.type_to_tag_map[field.field_type]
                ]
            )
        else:
            reporter.report_warning(
                title="Failed to Map View Field Type",
                message=f"Failed to map view field type {field.field_type}. Won't emit tags for measure and dimension",
            )

        # Add group_label as tags if present
        if field.group_label:
            schema_field_tags.append(
                TagAssociationClass(
                    tag=builder.make_tag_urn(f"looker:group_label:{field.group_label}")
                )
            )

        if schema_field_tags:
            return GlobalTagsClass(tags=schema_field_tags)

        return None

    @staticmethod
    def get_tag_mces() -> Iterable[MetadataChangeEvent]:
        # Emit tag MCEs for measures and dimensions:
        return [
            LookerUtil._get_tag_mce_for_urn(tag_urn)
            for tag_urn in LookerUtil.tag_definitions
        ]

    @staticmethod
    def view_field_to_schema_field(
        field: ViewField,
        reporter: SourceReport,
        tag_measures_and_dimensions: bool = True,
    ) -> SchemaField:
        return SchemaField(
            fieldPath=field.name,
            type=LookerUtil.get_field_type(field.type),
            nativeDataType=field.type,
            label=field.label,
            description=(
                field.description
                if tag_measures_and_dimensions is True
                else f"{field.field_type.value}. {field.description}"
            ),
            globalTags=(
                LookerUtil._get_tags_from_field_type(field, reporter)
                if tag_measures_and_dimensions is True
                else None
            ),
            isPartOfKey=field.is_primary_key,
        )

    @staticmethod
    def _get_fields_and_primary_keys(
        view_fields: List[ViewField],
        reporter: SourceReport,
        tag_measures_and_dimensions: bool = True,
    ) -> Tuple[List[SchemaField], List[str]]:
        primary_keys: List = []
        fields = []
        for field in view_fields:
            schema_field = LookerUtil.view_field_to_schema_field(
                field, reporter, tag_measures_and_dimensions
            )
            fields.append(schema_field)
            if field.is_primary_key:
                primary_keys.append(schema_field.fieldPath)
        return fields, primary_keys

    @staticmethod
    def _display_name(name: str) -> str:
        """Returns a display name that corresponds to the Looker conventions"""
        return name.replace("_", " ").title() if name else name

    @staticmethod
    def create_query_request(q: dict, limit: Optional[str] = None) -> WriteQuery:
        return WriteQuery(
            model=q["model"],
            view=q["view"],
            fields=q.get("fields"),
            filters=q.get("filters"),
            filter_expression=q.get("filter_expressions"),
            sorts=q.get("sorts"),
            limit=q.get("limit") or limit,
            column_limit=q.get("column_limit"),
            vis_config={"type": "looker_column"},
            filter_config=q.get("filter_config"),
            query_timezone="UTC",
        )


@dataclass
class LookerExplore:
    name: str
    model_name: str
    project_name: Optional[str] = None
    label: Optional[str] = None
    description: Optional[str] = None
    upstream_views: Optional[List[ProjectInclude]] = (
        None  # captures the view name(s) this explore is derived from
    )
    upstream_views_file_path: Dict[str, Optional[str]] = dataclasses_field(
        default_factory=dict
    )  # view_name is key and file_path is value. A single file may contains multiple views
    joins: Optional[List[str]] = None
    fields: Optional[List[ViewField]] = None  # the fields exposed in this explore
    source_file: Optional[str] = None
    tags: List[str] = dataclasses_field(default_factory=list)

    @validator("name")
    def remove_quotes(cls, v):
        # Sanitize the name.
        v = v.replace('"', "").replace("`", "")
        return v

    @staticmethod
    def _get_fields_from_sql_equality(sql_fragment: str) -> List[str]:
        field_match = re.compile(r"\${([^}]+)}")
        return field_match.findall(sql_fragment)

    @classmethod
    def from_dict(
        cls,
        model_name: str,
        dict: Dict,
        resolved_includes: List[ProjectInclude],
        looker_viewfile_loader: LookerViewFileLoader,
        reporter: LookMLSourceReport,
        model_explores_map: Dict[str, dict],
    ) -> "LookerExplore":
        view_names: Set[str] = set()
        joins = None
        assert "name" in dict, "Explore doesn't have a name field, this isn't allowed"

        # The view name that the explore refers to is resolved in the following order of priority:
        # 1. view_name: https://cloud.google.com/looker/docs/reference/param-explore-view-name
        # 2. from: https://cloud.google.com/looker/docs/reference/param-explore-from
        # 3. default to the name of the explore
        view_names.add(dict.get("view_name") or dict.get("from") or dict["name"])

        if dict.get("joins", {}) != {}:
            # additionally for join-based explores, pull in the linked views
            assert "joins" in dict
            for join in dict["joins"]:
                join_from = join.get("from")
                view_names.add(join_from or join["name"])
                sql_on = join.get("sql_on", None)
                if sql_on is not None:
                    fields = cls._get_fields_from_sql_equality(sql_on)
                    joins = fields

        upstream_views: List[ProjectInclude] = []
        # create the list of extended explores
        extends = list(
            itertools.chain.from_iterable(
                dict.get("extends", dict.get("extends__all", []))
            )
        )
        if extends:
            for extended_explore in extends:
                if extended_explore in model_explores_map:
                    parsed_explore = LookerExplore.from_dict(
                        model_name,
                        model_explores_map[extended_explore],
                        resolved_includes,
                        looker_viewfile_loader,
                        reporter,
                        model_explores_map,
                    )
                    upstream_views.extend(parsed_explore.upstream_views or [])
                else:
                    logger.warning(
                        f"Could not find extended explore {extended_explore} for explore {dict['name']} in model {model_name}"
                    )
        else:
            # we only fallback to the view_names list if this is not an extended explore
            for view_name in view_names:
                info = find_view_from_resolved_includes(
                    None,
                    resolved_includes,
                    looker_viewfile_loader,
                    view_name,
                    reporter,
                )
                if not info:
                    logger.warning(
                        f"Could not resolve view {view_name} for explore {dict['name']} in model {model_name}"
                    )
                else:
                    logger.debug(
                        f"LookerExplore.from_dict adding upstream view for explore '{dict['name']}' (model='{model_name}'): view_name='{view_name}', info[0].project='{info[0].project}'"
                    )
                    upstream_views.append(
                        ProjectInclude(project=info[0].project, include=view_name)
                    )

        return LookerExplore(
            model_name=model_name,
            name=dict["name"],
            label=dict.get("label"),
            description=dict.get("description"),
            upstream_views=upstream_views,
            joins=joins,
            # This method is getting called from lookml_source's get_internal_workunits method
            # & upstream_views_file_path is not in use in that code flow
            upstream_views_file_path={},
            tags=cast(List, dict.get("tags")) if dict.get("tags") is not None else [],
        )

    @classmethod
    def from_api(  # noqa: C901
        cls,
        model: str,
        explore_name: str,
        client: LookerAPI,
        reporter: SourceReport,
        source_config: LookerDashboardSourceConfig,
    ) -> Optional["LookerExplore"]:
        try:
            explore = client.lookml_model_explore(model, explore_name)

            views: Set[str] = set()
            lkml_fields: List[LookmlModelExploreField] = (
                explore_field_set_to_lkml_fields(explore)
            )

            if explore.view_name is not None and explore.view_name != explore.name:
                # explore is not named after a view and is instead using a from field, which is modeled as view_name.
                aliased_explore = True
                views.add(explore.view_name)
            else:
                # otherwise, the explore name is a view, so add it to the set.
                aliased_explore = False
                if explore.name is not None:
                    views.add(explore.name)

            if explore.joins is not None and explore.joins != []:
                join_to_orig_name_map = {}
                potential_views = []
                for e in explore.joins:
                    if e.from_ is not None:
                        potential_views.append(e.from_)
                        join_to_orig_name_map[e.name] = e.from_
                    elif e.name is not None:
                        potential_views.append(e.name)
                for e_join in [
                    e for e in explore.joins if e.dependent_fields is not None
                ]:
                    assert e_join.dependent_fields is not None
                    for field_name in e_join.dependent_fields:
                        try:
                            view_name = LookerUtil._extract_view_from_field(field_name)
                            orig_name = join_to_orig_name_map.get(e_join.name)
                            if orig_name is not None:
                                view_name = orig_name
                            potential_views.append(view_name)
                        except AssertionError:
                            reporter.report_warning(
                                title="Missing View Name",
                                message="The field was not prefixed by a view name. This can happen when the field references another dynamic field.",
                                context=field_name,
                            )
                            continue

                for view_name in potential_views:
                    if (view_name == explore.name) and aliased_explore:
                        # if the explore is aliased, then the joins could be referring to views using the aliased name.
                        # this needs to be corrected by switching to the actual view name of the explore
                        assert explore.view_name is not None
                        view_name = explore.view_name
                    views.add(view_name)

            view_fields: List[ViewField] = []
            field_name_vs_raw_explore_field: Dict = {}

            if explore.fields is not None:
                if explore.fields.dimensions is not None:
                    for dim_field in explore.fields.dimensions:
                        if dim_field.name is None:
                            continue
                        else:
                            field_name_vs_raw_explore_field[dim_field.name] = dim_field

                            view_fields.append(
                                ViewField(
                                    name=dim_field.name,
                                    label=dim_field.label_short,
                                    description=(
                                        dim_field.description
                                        if dim_field.description
                                        else ""
                                    ),
                                    type=(
                                        dim_field.type
                                        if dim_field.type is not None
                                        else ""
                                    ),
                                    field_type=(
                                        ViewFieldType.DIMENSION_GROUP
                                        if dim_field.dimension_group is not None
                                        else ViewFieldType.DIMENSION
                                    ),
                                    project_name=LookerUtil.extract_project_name_from_source_file(
                                        dim_field.source_file
                                    ),
                                    view_name=LookerUtil.extract_view_name_from_lookml_model_explore_field(
                                        dim_field
                                    ),
                                    is_primary_key=(
                                        dim_field.primary_key
                                        if dim_field.primary_key
                                        else False
                                    ),
                                    upstream_fields=[],
                                    group_label=dim_field.field_group_label,
                                )
                            )
                if explore.fields.measures is not None:
                    for measure_field in explore.fields.measures:
                        if measure_field.name is None:
                            continue
                        else:
                            field_name_vs_raw_explore_field[measure_field.name] = (
                                measure_field
                            )

                            view_fields.append(
                                ViewField(
                                    name=measure_field.name,
                                    label=measure_field.label_short,
                                    description=(
                                        measure_field.description
                                        if measure_field.description
                                        else ""
                                    ),
                                    type=(
                                        measure_field.type
                                        if measure_field.type is not None
                                        else ""
                                    ),
                                    field_type=ViewFieldType.MEASURE,
                                    project_name=LookerUtil.extract_project_name_from_source_file(
                                        measure_field.source_file
                                    ),
                                    view_name=LookerUtil.extract_view_name_from_lookml_model_explore_field(
                                        measure_field
                                    ),
                                    is_primary_key=(
                                        measure_field.primary_key
                                        if measure_field.primary_key
                                        else False
                                    ),
                                    upstream_fields=[],
                                    group_label=measure_field.field_group_label,
                                )
                            )

            view_project_map: Dict[str, str] = create_view_project_map(
                view_fields,
                explore_primary_view=explore.view_name,
                explore_project_name=explore.project_name,
            )
            if view_project_map:
                logger.debug(f"views and their projects: {view_project_map}")

            upstream_views_file_path: Dict[str, Optional[str]] = (
                create_upstream_views_file_path_map(
                    lkml_fields=lkml_fields,
                    view_names=views,
                )
            )
            if upstream_views_file_path:
                logger.debug(f"views and their file-paths: {upstream_views_file_path}")

            # form upstream of fields as all information is now available
            for view_field in view_fields:
                measure_upstream_field: ExploreUpstreamViewField = (
                    ExploreUpstreamViewField(
                        explore=explore,
                        field=field_name_vs_raw_explore_field[view_field.name],
                    )
                )

                assert explore.project_name is not None

                column_ref: Optional[ColumnRef] = measure_upstream_field.upstream(
                    view_project_map=view_project_map,
                    explore_project_name=explore.project_name,
                    model_name=model,
                    upstream_views_file_path=upstream_views_file_path,
                    config=source_config,
                )
                view_field.upstream_fields = (
                    [column_ref] if column_ref is not None else []
                )

            looker_explore = cls(
                name=explore_name,
                model_name=model,
                project_name=explore.project_name,
                label=explore.label,
                description=explore.description,
                fields=view_fields,
                upstream_views=list(
                    ProjectInclude(
                        project=view_project_map.get(view_name, BASE_PROJECT_NAME),
                        include=view_name,
                    )
                    for view_name in views
                ),
                upstream_views_file_path=upstream_views_file_path,
                source_file=explore.source_file,
                tags=list(explore.tags) if explore.tags is not None else [],
            )
            logger.debug(f"Created LookerExplore from API: {looker_explore}")
            return looker_explore
        except SDKError as e:
            if "<title>Looker Not Found (404)</title>" in str(e):
                logger.info(
                    f"Explore {explore_name} in model {model} is referred to, but missing. Continuing..."
                )
            else:
                logger.warning(
                    f"Failed to extract explore {explore_name} from model {model}: {e}"
                )
        except DeserializeError as e:
            reporter.warning(
                title="Failed to fetch explore from the Looker API",
                message=(
                    "An error occurred while extracting the explore from the model. "
                    "Please check the explore and model configurations."
                ),
                context=f"Explore: {explore_name}, Model: {model}",
                exc=e,
            )
        except AssertionError:
            reporter.report_warning(
                title="Unable to find Views",
                message="Encountered exception while attempting to find dependent views for this chart",
                context=f"Explore: {explore_name}, Mode: {model}, Views: {views}",
            )
        return None

    def get_mapping(self, config: LookerCommonConfig) -> NamingPatternMapping:
        return NamingPatternMapping(
            platform=config.platform_name,
            project=self.project_name,  # type: ignore
            model=self.model_name,
            name=self.name,
            env=config.env.lower(),
        )

    def get_explore_urn(self, config: LookerCommonConfig) -> str:
        dataset_name = config.explore_naming_pattern.replace_variables(
            self.get_mapping(config)
        )
        logger.debug(
            f"Generated dataset_name={dataset_name} for explore with model_name={self.model_name}, name={self.name}"
        )

        return builder.make_dataset_urn_with_platform_instance(
            platform=config.platform_name,
            name=dataset_name,
            platform_instance=config.platform_instance,
            env=config.env,
        )

    def get_explore_browse_path(self, config: LookerCommonConfig) -> str:
        browse_path = config.explore_browse_pattern.replace_variables(
            self.get_mapping(config)
        )
        return browse_path

    def _get_url(self, base_url):
        base_url = remove_port_from_url(base_url)
        return f"{base_url}/explore/{self.model_name}/{self.name}"

    def _get_embed_url(self, base_url: str) -> str:
        base_url = remove_port_from_url(base_url)
        return f"{base_url}/embed/explore/{self.model_name}/{self.name}"

    def _to_metadata_events(
        self,
        config: LookerCommonConfig,
        reporter: SourceReport,
        base_url: str,
        extract_embed_urls: bool,
    ) -> Dataset:
        """
        Generate a Dataset metadata event for this Looker Explore.

        Only generates datasets for explores that contain FROM clauses and do NOT contain joins.
        Passthrough explores and joins are handled via lineage and do not need additional nodes.
        """
        upstream_lineage = None
        view_name_to_urn_map: Dict[str, str] = {}

        if self.upstream_views is not None:
            assert self.project_name is not None
            upstreams: list[UpstreamClass] = []
            observed_lineage_ts = datetime.datetime.now(tz=datetime.timezone.utc)

            for view_ref in sorted(self.upstream_views):
                # set file_path to ViewFieldType.UNKNOWN if file_path is not available to keep backward compatibility
                # if we raise error on file_path equal to None then existing test-cases will fail as mock data
                # doesn't have required attributes.
                file_path: str = (
                    cast(str, self.upstream_views_file_path[view_ref.include])
                    if self.upstream_views_file_path.get(view_ref.include) is not None
                    else ViewFieldValue.NOT_AVAILABLE.value
                )

                view_urn = LookerViewId(
                    project_name=(
                        view_ref.project
                        if view_ref.project != BASE_PROJECT_NAME
                        else self.project_name
                    ),
                    model_name=self.model_name,
                    view_name=view_ref.include,
                    file_path=file_path,
                ).get_urn(config)

                upstreams.append(
                    UpstreamClass(
                        dataset=view_urn,
                        type=DatasetLineageTypeClass.VIEW,
                        auditStamp=AuditStamp(
                            time=int(observed_lineage_ts.timestamp() * 1000),
                            actor=CORPUSER_DATAHUB,
                        ),
                    )
                )
                view_name_to_urn_map[view_ref.include] = view_urn

            fine_grained_lineages: list[FineGrainedLineageClass] = []
            if config.extract_column_level_lineage:
                for field in self.fields or []:
                    # Skip creating fine-grained lineage for empty field names to prevent invalid schema field URNs
                    if not field.name or not field.name.strip():
                        logger.warning(
                            f"Skipping fine-grained lineage for field with empty name in explore '{self.name}'"
                        )
                        continue

                    for upstream_column_ref in field.upstream_fields:
                        # Skip creating fine-grained lineage for empty column names to prevent invalid schema field URNs
                        if (
                            not upstream_column_ref.column
                            or not upstream_column_ref.column.strip()
                        ):
                            logger.warning(
                                f"Skipping some fine-grained lineage for field '{field.name}' in explore '{self.name}' "
                                f"due to empty upstream column name in table '{upstream_column_ref.table}'"
                            )
                            continue

                        fine_grained_lineages.append(
                            FineGrainedLineageClass(
                                upstreamType=FineGrainedLineageUpstreamType.FIELD_SET,
                                downstreamType=FineGrainedLineageDownstreamType.FIELD,
                                upstreams=[
                                    builder.make_schema_field_urn(
                                        upstream_column_ref.table,
                                        upstream_column_ref.column,
                                    )
                                ],
                                downstreams=[
                                    builder.make_schema_field_urn(
                                        self.get_explore_urn(config), field.name
                                    )
                                ],
                            )
                        )

            upstream_lineage = UpstreamLineage(
                upstreams=upstreams,
                fineGrainedLineages=fine_grained_lineages or None,
            )

        schema_metadata = None
        if self.fields is not None:
            schema_metadata = LookerUtil._get_schema(
                platform_name=config.platform_name,
                schema_name=self.name,
                view_fields=self.fields,
                reporter=reporter,
            )

        extra_aspects: List[Union[GlobalTagsClass, EmbedClass]] = []

        explore_tag_urns: List[TagUrn] = [TagUrn(tag) for tag in self.tags]
        if extract_embed_urls:
            extra_aspects.append(EmbedClass(renderUrl=self._get_embed_url(base_url)))

        custom_properties: Dict[str, Optional[str]] = {
            "project": self.project_name,
            "model": self.model_name,
            "looker.explore.label": self.label,
            "looker.explore.name": self.name,
            "looker.explore.file": self.source_file,
        }

        return Dataset(
            platform=config.platform_name,
            name=config.explore_naming_pattern.replace_variables(
                self.get_mapping(config)
            ),
            display_name=str(self.label)
            if self.label
            else LookerUtil._display_name(self.name),
            description=self.description,
            subtype=DatasetSubTypes.LOOKER_EXPLORE,
            env=config.env,
            platform_instance=config.platform_instance,
            custom_properties={
                k: str(v) for k, v in custom_properties.items() if v is not None
            },
            external_url=self._get_url(base_url),
            upstreams=upstream_lineage,
            schema=schema_metadata,
            parent_container=[
                "Explore",
                gen_model_key(config, self.model_name).as_urn(),
            ],
            tags=explore_tag_urns if explore_tag_urns else None,
            extra_aspects=extra_aspects,
        )


def gen_project_key(config: LookerCommonConfig, project_name: str) -> LookMLProjectKey:
    return LookMLProjectKey(
        platform=config.platform_name,
        instance=config.platform_instance,
        env=config.env,
        project_name=project_name,
    )


def gen_model_key(config: LookerCommonConfig, model_name: str) -> LookMLModelKey:
    return LookMLModelKey(
        platform=config.platform_name,
        instance=config.platform_instance,
        env=config.env,
        model_name=model_name,
    )


class LookerExploreRegistry:
    """A LRU caching registry of Looker Explores"""

    def __init__(
        self,
        looker_api: LookerAPI,
        report: SourceReport,
        source_config: LookerDashboardSourceConfig,
    ):
        self.client = looker_api
        self.report = report
        self.source_config = source_config

    @lru_cache(maxsize=200)
    def get_explore(self, model: str, explore: str) -> Optional[LookerExplore]:
        logger.debug(f"Retrieving explore: model={model}, explore={explore}")
        looker_explore = LookerExplore.from_api(
            model,
            explore,
            self.client,
            self.report,
            self.source_config,
        )
        if looker_explore is not None:
            logger.debug(
                f"Found explore with model_name={looker_explore.model_name}, name={looker_explore.name}"
            )
        else:
            logger.debug(f"No explore found for model={model}, explore={explore}")
        return looker_explore

    def compute_stats(self) -> Dict:
        return {
            "cache_info": self.get_explore.cache_info(),
        }


class StageLatency(Report):
    name: str
    start_time: Optional[datetime.datetime]
    end_time: Optional[datetime.datetime] = None

    def __init__(self, name: str, start_time: datetime.datetime):
        self.name = name
        self.start_time = start_time

    def compute_stats(self) -> None:
        if self.end_time and self.start_time:
            self.latency = self.end_time - self.start_time
            # clear out start and end times to keep logs clean
            self.start_time = None
            self.end_time = None


@dataclass
class LookerDashboardSourceReport(StaleEntityRemovalSourceReport):
    total_dashboards: int = 0
    dashboards_scanned: int = 0
    looks_scanned: int = 0
    filtered_dashboards: LossyList[str] = dataclasses_field(default_factory=LossyList)
    filtered_looks: LossyList[str] = dataclasses_field(default_factory=LossyList)
    dashboards_scanned_for_usage: int = 0
    charts_scanned_for_usage: int = 0
    charts_with_activity: LossySet[str] = dataclasses_field(default_factory=LossySet)
    accessed_dashboards: int = 0
    dashboards_with_activity: LossySet[str] = dataclasses_field(
        default_factory=LossySet
    )

    # Entities that don't seem to exist, so we don't emit usage aspects for them despite having usage data
    dashboards_skipped_for_usage: LossySet[str] = dataclasses_field(
        default_factory=LossySet
    )
    charts_skipped_for_usage: LossySet[str] = dataclasses_field(
        default_factory=LossySet
    )

    stage_latency: List[StageLatency] = dataclasses_field(default_factory=list)
    _looker_explore_registry: Optional[LookerExploreRegistry] = None
    total_explores: int = 0
    explores_scanned: int = 0

    resolved_user_ids: int = 0
    email_ids_missing: int = 0  # resolved users with missing email addresses
    looker_user_count: int = 0

    _looker_api: Optional[LookerAPI] = None
    query_latency: Dict[str, datetime.timedelta] = dataclasses_field(
        default_factory=dict
    )
    user_resolution_latency: Dict[str, datetime.timedelta] = dataclasses_field(
        default_factory=dict
    )

    def report_total_dashboards(self, total_dashboards: int) -> None:
        self.total_dashboards = total_dashboards

    def report_dashboards_scanned(self) -> None:
        self.dashboards_scanned += 1

    def report_charts_scanned(self) -> None:
        self.looks_scanned += 1

    def report_dashboards_dropped(self, model: str) -> None:
        self.filtered_dashboards.append(model)

    def report_charts_dropped(self, view: str) -> None:
        self.filtered_looks.append(view)

    def report_dashboards_scanned_for_usage(self, num_dashboards: int) -> None:
        self.dashboards_scanned_for_usage += num_dashboards

    def report_charts_scanned_for_usage(self, num_charts: int) -> None:
        self.charts_scanned_for_usage += num_charts

    def report_upstream_latency(
        self, start_time: datetime.datetime, end_time: datetime.datetime
    ) -> None:
        # recording total combined latency is not very useful, keeping this method as a placeholder
        # for future implementation of min / max / percentiles etc.
        pass

    def report_query_latency(
        self, query_type: str, latency: datetime.timedelta
    ) -> None:
        self.query_latency[query_type] = latency

    def report_user_resolution_latency(
        self, generator_type: str, latency: datetime.timedelta
    ) -> None:
        self.user_resolution_latency[generator_type] = latency

    def report_stage_start(self, stage_name: str) -> None:
        self.stage_latency.append(
            StageLatency(name=stage_name, start_time=datetime.datetime.now())
        )

    def report_stage_end(self, stage_name: str) -> None:
        if self.stage_latency[-1].name == stage_name:
            self.stage_latency[-1].end_time = datetime.datetime.now()

    @contextmanager
    def report_stage(self, stage_name: str) -> Iterator[None]:
        try:
            self.report_stage_start(stage_name)
            yield
        finally:
            self.report_stage_end(stage_name)

    def compute_stats(self) -> None:
        if self.total_dashboards:
            self.dashboard_process_percentage_completion = round(
                100 * self.dashboards_scanned / self.total_dashboards, 2
            )
        if self._looker_explore_registry:
            self.explore_registry_stats = self._looker_explore_registry.compute_stats()

        if self.total_explores:
            self.explores_process_percentage_completion = round(
                100 * self.explores_scanned / self.total_explores, 2
            )

        if self._looker_api:
            self.looker_api_stats = self._looker_api.compute_stats()
        return super().compute_stats()


@dataclass
class LookerUser:
    id: str
    email: Optional[str]
    display_name: Optional[str]
    first_name: Optional[str]
    last_name: Optional[str]

    @classmethod
    def create_looker_user(cls, raw_user: User) -> "LookerUser":
        assert raw_user.id is not None
        return LookerUser(
            raw_user.id,
            raw_user.email,
            raw_user.display_name,
            raw_user.first_name,
            raw_user.last_name,
        )

    def get_urn(self, strip_user_ids_from_email: bool) -> Optional[str]:
        user = self.email
        if user and strip_user_ids_from_email:
            user = user.split("@")[0]

        if not user:
            return None
        return builder.make_user_urn(user)


@dataclass
class InputFieldElement:
    name: str
    view_field: Optional[ViewField]
    model: str = ""
    explore: str = ""


@dataclass
class LookerDashboardElement:
    id: str
    title: str
    query_slug: str
    upstream_explores: List[LookerExplore]
    look_id: Optional[str]
    type: Optional[str] = None
    description: Optional[str] = None
    input_fields: Optional[List[InputFieldElement]] = None
    folder_path: Optional[str] = None  # for independent looks.
    folder: Optional[LookerFolder] = None
    owner: Optional[LookerUser] = None

    def url(self, base_url: str) -> str:
        # A dashboard element can use a look or just a raw query against an explore
        base_url = remove_port_from_url(base_url)
        if self.look_id is not None:
            return f"{base_url}/looks/{self.look_id}"
        else:
            return f"{base_url}/x/{self.query_slug}"

    def embed_url(self, base_url: str) -> Optional[str]:
        # A dashboard element can use a look or just a raw query against an explore
        base_url = remove_port_from_url(base_url)
        if self.look_id is not None:
            return f"{base_url}/embed/looks/{self.look_id}"
        else:
            # No embeddable URL
            return None

    def get_urn_element_id(self):
        # A dashboard element can use a look or just a raw query against an explore
        return f"dashboard_elements.{self.id}"

    def get_view_urns(self, config: LookerCommonConfig) -> List[str]:
        return [v.get_explore_urn(config) for v in self.upstream_explores]


# These function will avoid to create LookerDashboard object to get the Looker Dashboard urn id part
def get_urn_looker_dashboard_id(id_: str) -> str:
    return f"dashboards.{id_}"


def get_urn_looker_element_id(id_: str) -> str:
    return f"dashboard_elements.{id_}"


@dataclass
class LookerDashboard:
    id: str
    title: str
    dashboard_elements: List[LookerDashboardElement]
    created_at: Optional[datetime.datetime]
    description: Optional[str] = None
    folder_path: Optional[str] = None
    folder: Optional[LookerFolder] = None
    is_deleted: bool = False
    is_hidden: bool = False
    owner: Optional[LookerUser] = None
    strip_user_ids_from_email: Optional[bool] = True
    last_updated_at: Optional[datetime.datetime] = None
    last_updated_by: Optional[LookerUser] = None
    deleted_at: Optional[datetime.datetime] = None
    deleted_by: Optional[LookerUser] = None
    favorite_count: Optional[int] = None
    view_count: Optional[int] = None
    last_viewed_at: Optional[datetime.datetime] = None

    def url(self, base_url):
        base_url = remove_port_from_url(base_url)
        return f"{base_url}/dashboards/{self.id}"

    def embed_url(self, base_url: str) -> str:
        base_url = remove_port_from_url(base_url)
        return f"{base_url}/embed/dashboards/{self.id}"

    def get_urn_dashboard_id(self):
        return get_urn_looker_dashboard_id(self.id)


class LookerUserRegistry:
    looker_api_wrapper: LookerAPI
    fields: str = ",".join(["id", "email", "display_name", "first_name", "last_name"])
    _user_cache: Dict[str, LookerUser] = {}

    def __init__(self, looker_api: LookerAPI, report: LookerDashboardSourceReport):
        self.looker_api_wrapper = looker_api
        self.report = report
        self._initialize_user_cache()

    def _initialize_user_cache(self) -> None:
        raw_users: Sequence[User] = self.looker_api_wrapper.all_users(
            user_fields=self.fields
        )

        for raw_user in raw_users:
            looker_user = LookerUser.create_looker_user(raw_user)
            self._user_cache[str(looker_user.id)] = looker_user

    def get_by_id(self, id_: str) -> Optional[LookerUser]:
        if not id_:
            return None

        logger.debug(f"Will get user {id_}")

        if str(id_) in self._user_cache:
            return self._user_cache.get(str(id_))

        raw_user: Optional[User] = self.looker_api_wrapper.get_user(
            str(id_), user_fields=self.fields
        )
        if raw_user is None:
            return None

        looker_user = LookerUser.create_looker_user(raw_user)
        return looker_user

    def to_platform_resource(
        self, platform_instance: Optional[str]
    ) -> Iterable[MetadataChangeProposalWrapper]:
        try:
            platform_resource_key = PlatformResourceKey(
                platform=LOOKER,
                resource_type="USER_ID_MAPPING",
                platform_instance=platform_instance,
                primary_key="",
            )

            # Extract user email mappings.
            # Sort it to ensure the order is deterministic.
            user_email_cache = {
                user_id: user.email
                for user_id, user in sorted(self._user_cache.items())
                if user.email
            }

            platform_resource = PlatformResource.create(
                key=platform_resource_key,
                value=user_email_cache,
            )

            self.report.looker_user_count = len(user_email_cache)
            yield from platform_resource.to_mcps()

        except Exception as exc:
            self.report.warning(
                message="Failed to generate platform resource for looker id mappings",
                exc=exc,
            )



from typing import Optional

from pydantic import Field, SecretStr

from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StatefulIngestionConfigBase,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulLineageConfigMixin,
    StatefulUsageConfigMixin,
)


class SnaplogicConfig(
    StatefulIngestionConfigBase, StatefulLineageConfigMixin, StatefulUsageConfigMixin
):
    platform: str = "Snaplogic"
    username: str = Field(description="Username")
    password: SecretStr = Field(description="Password")
    base_url: str = Field(
        default="https://elastic.snaplogic.com",
        description="Url to your Snaplogic instance: `https://elastic.snaplogic.com`, or similar. Used for making API calls to Snaplogic.",
    )
    org_name: str = Field(description="Organization name from Snaplogic instance")
    namespace_mapping: dict = Field(
        default={}, description="Mapping of namespaces to platform instances"
    )
    case_insensitive_namespaces: list = Field(
        default=[],
        description="List of namespaces that should be treated as case insensitive",
    )
    create_non_snaplogic_datasets: bool = Field(
        default=False,
        description="Whether to create datasets for non-Snaplogic datasets (e.g., databases, S3, etc.)",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = None
